# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

DAYS_IN_YEAR = 365.25
HOURS_IN_DAY = 24
SECONDS_IN_HOUR = 3600

class Yield():
             
    def __seconds_per_year(self):
        return DAYS_IN_YEAR*HOURS_IN_DAY*SECONDS_IN_HOUR

    def apply(self, A0, delta_t, apy):  
        multiplier = 1 + apy
        freq = self.__seconds_per_year()
        A1 = A0*(multiplier)**(delta_t/freq)
        return A1-A0


from math import sqrt
from ...cpt.quote import LPQuote
from ...utils.tools.v3 import TickMath

class UniswapImpLoss:
    def __init__(self, lp, lp_init_amt, lwr_tick = None, upr_tick = None):
        self.lp = lp
        self.lp_init = lp_init_amt
        self.x_tkn_init = self._calc_dx(lp_init_amt, upr_tick)
        self.y_tkn_init = self._calc_dy(lp_init_amt, lwr_tick)

    def current_position_value(self, tkn, lwr_tick = None, upr_tick = None):
        """Calculate the current value of the LP position in terms of input token.""" 
        return LPQuote(False).get_amount_from_lp(self.lp, tkn, self.lp_init, lwr_tick, upr_tick)

    def hold_value(self, tkn):
        """Calculate the value if initial tokens were held."""
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        if(tkn.token_name == self.lp.token0):
            current_price = self.lp.get_price(tokens[self.lp.token1])
            val = self.y_tkn_init*current_price + self.x_tkn_init
        elif(tkn.token_name == self.lp.token1):   
            current_price = self.lp.get_price(tokens[self.lp.token0])
            val = self.x_tkn_init * current_price + self.y_tkn_init
        return val 

    def get_init_amt(self, tkn):
        if(tkn.token_name == self.lp.token0):
            return self.x_tkn_init
        elif(tkn.token_name == self.lp.token1):  
            return self.y_tkn_init

    def apply(self, lwr_tick = None, upr_tick = None, fees = False):

        if(fees):
            """Calculate returns based current position."""
            tokens = self.lp.factory.token_from_exchange[self.lp.name]
            x_tkn = tokens[self.lp.token0]
            v_hold = self.hold_value(x_tkn)
            v_pos = self.current_position_value(x_tkn, lwr_tick, upr_tick)
            iloss = (v_pos - v_hold)/v_hold
        else:
            """Calculate IL based on price ratio."""
            tokens = self.lp.factory.token_from_exchange[self.lp.name]
            x_tkn = tokens[self.lp.token0]
            y_tkn = tokens[self.lp.token1]
            initial_price = self.y_tkn_init / self.x_tkn_init
            current_price = self.lp.get_price(x_tkn)
            alpha = current_price / initial_price
            if(self.lp.version == 'V2'):
                iloss = self.calc_iloss(alpha)
            elif(self.lp.version == 'V3'):    
                r = self.calc_price_range(lwr_tick, upr_tick)
                iloss = self.calc_iloss(alpha, r)       

        return iloss

    def calc_iloss(self, alpha, r = None):
        if(r == None):
            return (2 * sqrt(alpha)) / (1 + alpha) - 1
        else:
            iloss = (2 * sqrt(alpha)) / (1 + alpha) - 1
            scale =  sqrt(r)/(sqrt(r)-1)
            return scale*iloss 

    def calc_price_range(self, lwr_tick, upr_tick):
        Q96 = 2**96
        sqrtp_cur = self.lp.slot0.sqrtPriceX96/Q96
        sqrtp_pb = TickMath.getSqrtRatioAtTick(upr_tick)/Q96
        sqrtp_pa = TickMath.getSqrtRatioAtTick(lwr_tick)/Q96
        ra = sqrtp_pa**2/sqrtp_cur**2
        rb = sqrtp_pb**2/sqrtp_cur**2    
        return sum([ra,rb])/2

    def _calc_dx(self, dL, upr_tick = None):
        if(self.lp.version == 'V2'):
            return self._calc_univ2_dx(dL)
        elif(self.lp.version == 'V3'): 
            return self._calc_univ3_dx(dL, upr_tick)

    def _calc_dy(self, dL, lwr_tick = None):
        if(self.lp.version == 'V2'):
            return self._calc_univ2_dy(dL)
        elif(self.lp.version == 'V3'): 
            return self._calc_univ3_dy(dL, lwr_tick)
        
    def _calc_univ2_dx(self, dL):
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        x_tkn = tokens[self.lp.token0]
        x = self.lp.get_reserve(x_tkn)
        L = self.lp.get_liquidity()
        return x*dL/L

    def _calc_univ2_dy(self, dL):
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        y_tkn = tokens[self.lp.token1]
        y = self.lp.get_reserve(y_tkn)
        L = self.lp.get_liquidity()
        return y*dL/L
    
    def _calc_univ3_dx(self, dL, upr_tick):
        Q96 = 2**96
        sqrtp_pb = TickMath.getSqrtRatioAtTick(upr_tick)/Q96
        sqrtp_cur = self.lp.slot0.sqrtPriceX96/Q96
        dPx = (1/sqrtp_cur - 1/sqrtp_pb)     
        return dL*dPx

    def _calc_univ3_dy(self, dL, lwr_tick):
        Q96 = 2**96
        sqrtp_pa = TickMath.getSqrtRatioAtTick(lwr_tick)/Q96
        sqrtp_cur = self.lp.slot0.sqrtPriceX96/Q96
        dPy = (sqrtp_cur - sqrtp_pa) 
        return dL*dPy


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from .ERC20 import ERC20

class IndexERC20(ERC20):
    
    """ Index ERC20 token

        Parameters
        ----------
        token_name : str
            Token name 
        token_addr : str
            Token address  
        token_total : float
            Token holdings 
    """  

    def __init__(self, name, addr, parent_tkn, parent_lp):
        self.token_name = name
        self.token_addr = addr
        self.parent_lp = parent_lp
        self.parent_tkn = parent_tkn
        self.token_supply = 0
        self.token_total = 0
        self.type = 'index'
        
        
    def rebase(self): 
        
        """ rebase

            Rebase token 
                
            Parameters
            -------
            _to : str
                user address   
            value : str
                reset token value                
        """         
        
        self.token_total = new_total
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from decimal import Decimal

class RoundFloat():
    
    def apply(self, num, trunc_ratio = 0.5):
        n_float = self.n_float_digits(num)
        trunc = int(n_float*trunc_ratio)
        return round(num, trunc)
    
    def n_float_digits(self, num):
        d = Decimal(str(num))
        return abs(d.as_tuple().exponent)


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np

class MaxDrop():

    """ Determine maximum percentage drop for any numerical array along with its
        starting and end points
    """   
    
    def __init__(self):
        self.__drop = None
        self.__pnt1 = None
        self.__pnt2 = None
    
    def apply(self, arr):
        
        """ apply

            Find maximum percentage drop

            Parameters
            ----------
            arr : numpy.array, shape (n_samples, 1)
                Numerical array 

            Returns
            -------
            x pnt, y pnt, drop : tuple, tuple, float
                Array of optimized leverage values              
        """        
        
        d_downs = self.calc_drawdowns(arr)
        arg_maxis = self.calc_arg_maximums(arr)
        maxis = self.calc_maximums(arr)
        idx = np.argmax(d_downs)
        self.__pnt1 = (arg_maxis[idx], maxis[idx])
        self.__pnt2 = (idx, arr[idx])
        self.__drop = self.calc_drop()
        return self.__pnt1, self.__pnt2, self.__drop  
 
    def calc(self):
        
        """ calc

            Calculate percentage drop given y_start and y_end values

            Returns
            -------
            percent drop : float
                Percentage drop             
        """         
        
        return (self.__pnt1[1]-self.__pnt2[1])/self.__pnt1[1]
    
    def get_pnt1(self):
        
        """ get_pnt1

            Get start point (x,y) of max percentage drop

            Returns
            -------
            self.__pnt1 : tuple
                Start point (x, y)            
        """         
        
        return self.__pnt1 
    
    def get_pnt2(self):
        
        """ get_pnt1

            Get end point (x,y) of max percentage drop

            Returns
            -------
            __pnt1 : tuple
                End point (x, y)            
        """         
        
        return self.__pnt2      
        
    def n_diff(self, arr, N):
        
        """ n_diff

            Generate N differenced time series
            
            Parameters
            ----------
            arr : numpy.array, shape (n_samples, 1)
                Numerical array  
            N : int
                Number of differences                 

            Returns
            -------
            arr_diff : numpy.array, shape (n_samples-1, 1)
                N differenced time series             
        """          
        
        arr = np.array(arr)
        return arr[N:] - arr[:-N]

    def calc_maximums(self, arr_in):  
        
        """ calc_maximums

            Generate monotone increasing array representing max events which increases at every 
            max event of input array
            
            Parameters
            ----------
            arr_in : numpy.array, shape (n_samples, 1)
                Numerical array  
             
            Returns
            -------
            maximums : numpy.array, shape (n_samples, 1)
                Generate monotone increasing array representing max events at every increase            
        """         
        
        N = len(arr_in)
        m = lambda arr, k: max(arr[:k]) if k > 0 else arr[0]   
        maximums = np.array([m(arr_in, k) for k in range(N)])
        return maximums 

    def calc_arg_maximums(self, arr_in):  
        
        """ calc_arg_maximums

            Generate monotone increasing array representing indices at max events which increases 
            at every max event of input array
            
            Parameters
            ----------
            arr_in : numpy.array, shape (n_samples, 1)
                Numerical array  
             
            Returns
            -------
            maximums : numpy.array, shape (n_samples, 1)
                Monotone increasing array representing max events at every increase            
        """          
        
        N = len(arr_in)
        arg_m = lambda arr, k: np.argmax(arr[:k]) if k > 0 else 0  
        arg_maximums = np.array([arg_m(arr_in, k) for k in range(N)])
        return arg_maximums 
    
    def calc_drawdowns(self, arr):
        
        """ calc_drawdowns

            Calculate running drawdowns from every max event of input array; running drawdown gets reset
            to zero when a new max event occurs
            
            Parameters
            ----------
            arr : numpy.array, shape (n_samples, 1)
                Numerical array  
             
            Returns
            -------
            running drawdowns : numpy.array, shape (n_samples, 1)
                Running drawdowns from every max event of input array           
        """      
    
        maxis = self.calc_maximums(arr)
        return (maxis - arr)/maxis

    def m_drawdown(self, arr):
        
        """ m_drawdown

            Calculate maximum drawdown from input array
            
            Parameters
            ----------
            arr : numpy.array, shape (n_samples, 1)
                Numerical array  
             
            Returns
            -------
            Maximum drawdown : float
                Maximum drawdown from input array           
        """         
        
        maxis = self.calc_maximums(arr)
        d_downs = (maxis - arr)/maxis
        idx = np.argmax(d_downs) 
        return d_downs[idx] 
        


# Copyright [2024] [Ian Moore, Bart Hofkin]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...utils.data import UniswapExchangeData
from ...utils.data import Chain0x
from ...erc import ERC20
from ...utils.client import API0x
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...cpt.factory import UniswapFactory
from ...cpt.quote import LPTokenQuote
from ...cpt.quote import IndexTokenQuote
from ...cpt.quote import TreeAmountQuote
from ...simulate import CorrectReserves
from ...process.deposit import SwapDeposit
from ...process.swap import WithdrawSwap
from ...process.swap import Swap

import time
import datetime
import random

USER_NM = 'user'

class QuantTerminal:
    
    """ 
        Quant Terminal Simulator class applies the 0x API to produce a mock Uniswap pool to 
        allow end-users to stress test the limitations of a Uniswap pool setup using live price 
        feeds from 0x API

        Parameters
        -----------------
        buy_token : str
            buy token contract address (ie, pulled from Chain0x data class)
        sell_token : str
            buy token contract address (ie, pulled from Chain0x data class)  
        time_window : float
            trial time window (ie, pulled from Chain0x data class, or passed in directly)  
        trade_bias : float
            Random trade bias where probability of bias to buy verses sell token postions is set to 50/50 by default 
            (ie, pulled from Chain0x data class, or passed in directly from slider in GUI) 
        td_model : TokenDeltaModel
            Token delta model is the non-deterministic model for incoming swap amounts; set 
            to Gamma distribution with paramaters set to scale = 1 and shape = 1 (by default)
        api : API0x
            Ox API class; unless otherwise specified, set to ETHEREUM chain and WETH/USDC trading pair by default
        
    """        
    
    STATE_ARB = 'arb'
    STATE_SWAP = 'swap'
    STATE_INIT = 'init_lp'
    STATE_INSTANTIATE = 'instantiate'
    
    def __init__(self, 
                 buy_token = None,
                 sell_token = None,                 
                 time_window = None, 
                 trade_bias = None,
                 td_model = None,
                 api = None
                ):
        
        self.buy_token = Chain0x().get_buy_token() if buy_token == None else buy_token
        self.sell_token = Chain0x().get_sell_token() if sell_token == None else sell_token
        self.time_window = Chain0x.time_window if time_window == None else time_window
        self.trade_bias = Chain0x.trade_bias if trade_bias == None else trade_bias        
        self.td_model = TokenDeltaModel(30) if td_model == None else td_model
        self.api = API0x(chain = Chain0x.ETHEREUM) if api == None else api 
        
        self.lp = None 
        self.arb = None
        self.x_tkn = None
        self.y_tkn = None
        self.time_init = None
        self.init_lp_invest = None
        self.state = self.STATE_INSTANTIATE
        self.remaining_sleep = self.time_window
        self.api_0x_data = None
        self.prev_api_0x_data = None
        
        self.x_redeem = None
        self.y_redeem = None
        
        self.x_trial_amt = 0
        self.y_trial_amt = 0        
        
        self.time_arb = None
        self.x_amt_arb = None
        self.y_amt_arb = None
        self.price_arb = None
        
        self.time_swap = None
        self.x_amt_swap = None
        self.y_amt_swap = None
        self.price_swap = None     
        self.amt_swap = None
        
    def get_lp(self):
        
        """ get_lp

            Get Uniswap exchange pool object   
                
            Returns
            -----------------
            lp : UniswapExchange
                Uniswap exchange pool object                 
        """          
        
        return self.lp
    
    def get_state(self):
        
        """ get_state

            Get pool's current state (ie, init_lp, arb, swap)
                
            Returns
            -----------------
            state : str
                Pool's current state (ie, init_lp, arb, swap)                 
        """         
        
        return self.state
    
    def get_swap_amt(self):
        
        """ get_swap_amt

            Get pool's current state (ie, init_lp, arb, swap)
                
            Returns
            -----------------
            amt_swap : float
                Pool's current state (ie, init_lp, arb, swap)                 
        """         
        
        return self.amt_swap  
    
    def get_x_redeem(self):
        
        """ get_x_redeem

            Get x token redemption amount from mock pool investment
                
            Returns
            -----------------
            x_redeem : float
                x token redemption amount from mock pool investment                  
        """           
        
        return self.x_redeem
    
    def get_y_redeem(self):
        
        """ get_y_redeem

            Get y token redemption amount from mock pool investment
                
            Returns
            -----------------
            y_redeem : float
                y token redemption amount from mock pool investment                  
        """          
        
        return self.y_redeem    
    
    def get_x_tkn(self):
        
        """ get_x_tkn

            Get x token from Uniswap mock pool
                
            Returns
            -----------------
            x_tkn : ERC20
                x token from mock pool                   
        """          
        
        return self.x_tkn    
    
    def get_y_tkn(self):
        
        """ get_y_tkn

            Get y token from Uniswap mock pool
                
            Returns
            -----------------
            y_tkn : ERC20
                y token from mock pool                   
        """            
        
        return self.y_tkn        
    
    def get_time_stamp(self, state):
        
        """ get_time_stamp

            Get datetime time stamp for a given pool state
                
            Parameters
            -----------------
            state : str
                Mock pool state (ie, arb, swap)  

            Returns
            -----------------
            time : datetime
                datetime timestamp                    
        """          
        
        if(state == self.STATE_SWAP):
            return self.time_swap 
        else:
            return self.time_arb    
    
    def get_lp_price(self, state):
        
        """ get_lp_price

            Get liquidity pool price for a given pool state
                
            Parameters
            -----------------
            state : str
                Mock pool state (ie, arb, swap)  

            Returns
            -----------------
            price : float
                LP price x/y (NOTE: returns x/y not y/x due to a nuance in SolveDeltas class)                      
        """            
        
        if(state == self.STATE_SWAP):
            return self.price_swap 
        else:
            return self.price_arb
    
    def get_x_reserve(self, state):
        
        """ get_x_reserve

            Get liquidity pool x reserve for a given pool state
                
            Parameters
            -----------------
            state : str
                Mock pool state (ie, arb, swap)  

            Returns
            -----------------
            x reserve : float
                Amount of x reserve in pool                   
        """          
        
        if(state == self.STATE_SWAP):
            return self.x_amt_swap 
        else:
            return self.x_amt_arb       
        
    def get_y_reserve(self, state):
        
        """ get_y_reserve

            Get liquidity pool y reserve for a given pool state
                
            Parameters
            -----------------
            state : str
                Mock pool state (ie, arb, swap)  

            Returns
            -----------------
            y reserve : float
                Amount of y reserve in pool                   
        """         
        
        if(state == self.STATE_SWAP):
            return self.y_amt_swap 
        else:
            return self.y_amt_arb 
        
    def get_usd_trial_volume(self):
        
        """ get_usd_trial_volume

            Get trade volume valued in stable token (eg. USDC) that passes through mock pool after a trial pass
            
            Returns
            -----------------
            volume : float
                Amount of total volume that passes through pool on both sides (ie, x and y) of the trade                  
        """           
        
        return TreeAmountQuote().get_tot_y(self.lp, self.x_trial_amt, self.y_trial_amt)
        
    def get_0x_data(self):    
        
        """ get_0x_data

            Get returned JSON structure after 0x API call 
                
            Returns
            -----------------
            0x api data : dictionary
                JSON structure after 0x api call                   
        """          
        
        return self.api_0x_data      
        
    def get_market_price(self): 
        
        """ get_market_price

            Get market price returned from 0x API call 
                
            Returns
            -----------------
            price : float
                price valued in x/y (NOTE: returns x/y not y/x due to a nuance in SolveDeltas class)                   
        """          
        
        return 1/float(self.api_0x_data['price']) if self.api_0x_data != None else -1
    
    def call_0x_api(self):    
        
        """ call_0x_api

            Call 0x API  
                
            Returns
            -----------------
            0x api data : dictionary
                JSON structured 0x api call data                 
        """   
        call_api = self.api.apply(self.sell_token, self.buy_token, Chain0x().get_api_sell_amount())   
        return call_api if bool(call_api) else self.prev_api_0x_data   
    
    def init_lp(self, init_x_tkn, x_tkn_nm = None, y_tkn_nm = None, init_x_invest = 1):
        
        """ init_lp

            Initialize mock liquidity pool
                
            Parameters
            -----------------
            init_x_tkn : float
                Initial x token amount
            x_tkn_nm : str
                x token name                
            y_tkn_nm : str
                y token name
            init_x_invest : float
                Initial x mock investment into pool (eg. if x tkn is WETH then init_x_invest = 1 would be 1 WETH)                             
        """          
        
        self.state = self.STATE_INIT
        self.api_0x_data = self.call_0x_api()
        p = self.get_market_price()
        x_tkn_amt = init_x_tkn
        y_tkn_amt = x_tkn_amt*p
        
        x_tkn_nm = Chain0x.buy_tkn_nm if x_tkn_nm == None else x_tkn_nm
        y_tkn_nm = Chain0x.sell_tkn_nm if y_tkn_nm == None else y_tkn_nm

        self.x_tkn = ERC20(x_tkn_nm, None)
        self.y_tkn = ERC20(y_tkn_nm, None)
        exchg_data = UniswapExchangeData(tkn0 = self.x_tkn, tkn1 = self.y_tkn, symbol="LP", address=None)

        factory = UniswapFactory("ETH pool factory", None)
        self.lp = factory.deploy(exchg_data)
        self.lp.add_liquidity(USER_NM, x_tkn_amt, y_tkn_amt, x_tkn_amt, y_tkn_amt) 
        self.take_x_position(init_x_invest)
                
        self.arb = CorrectReserves(self.lp, x0 = p)
        self.time_init = datetime.datetime.now()

     
    def trial(self):
        
        """ trial

            One trial pass includes: 1 random swap where amount is Gamma distributed which offsets pool price
            from the market price that the 0x API returns; this event is followed by an arbitrage to bring
            mock pool price in balance with the market price
                
            Returns
            -----------------
            p : float
                LP price x/y (NOTE: returns x/y not y/x due to a nuance in SolveDeltas class)                      
        """            
        
        self._reset_trial()
        self.api_0x_data = self.call_0x_api()
        p = self.get_market_price()
        remaining_sleep = self.time_window
        
        # STATE: RANDOM SWAP
        pause = random.uniform(0, remaining_sleep)
        time.sleep(pause)        
        self._random_swap(p)
        remaining_sleep -= pause
        
        # STATE: MARKET ARBITRAGE
        pause = random.uniform(0, remaining_sleep)
        time.sleep(pause) 
        self._market_arbitrage(p)
        remaining_sleep -= pause
        
        self._update_investment()
        self.prev_api_0x_data = self.api_0x_data
        time.sleep(remaining_sleep) 

        return p
    
    def take_x_position(self, x_invest):
        
        """ take_x_position

            Take mock position on pool; this is for the purpose of monitoring investment performance
            of trade over time within simulation
                
            Parameters
            -----------------
            x_invest : float
                Mock x position; eg. we invest 1 WETH into pool to monitor performance                   
        """          
        
        x_invest = x_invest if x_invest == 1 else x_invest
        self.init_lp_invest = LPTokenQuote().get_x(self.lp, x_invest)
        SwapDeposit().apply(self.lp, self.get_x_tkn(), USER_NM, x_invest)
        self._update_investment()    
        
    def _random_swap(self, p):   
        self.state = self.STATE_SWAP
        rnd_amt = self.td_model.delta()
        self.amt_swap = rnd_amt
        select_tkn = EventSelectionModel().bi_select(self.trade_bias)
        bias_factor = 1
        
        if(select_tkn == 0):
            out = Swap().apply(self.lp, self.x_tkn, USER_NM, rnd_amt) 
            self.x_trial_amt += rnd_amt
        else:
            out = Swap().apply(self.lp, self.y_tkn, USER_NM,  bias_factor*p*rnd_amt)  
            self.y_trial_amt += bias_factor*p*rnd_amt
            
        self.time_swap = datetime.datetime.now() 
        self.x_amt_swap = self.lp.get_reserve(self.arb.get_x_tkn())
        self.y_amt_swap = self.lp.get_reserve(self.arb.get_y_tkn())
        self.price_swap = self.lp.get_price(self.arb.get_x_tkn())        
        
    def _update_investment(self): 
        self.x_redeem = IndexTokenQuote().get_x(self.lp, self.init_lp_invest)
        self.y_redeem = IndexTokenQuote().get_y(self.lp, self.init_lp_invest)
        
    def _reset_trial(self):
        self.x_trial_amt = 0
        self.y_trial_amt = 0
            
    def _market_arbitrage(self, p):  
        self.state = self.STATE_ARB
        self.arb.apply(p)
        self.x_trial_amt += self.arb.get_swap_dx()
        self.y_trial_amt += self.arb.get_swap_dy()
        
        self.time_arb = datetime.datetime.now()
        self.x_amt_arb = self.lp.get_reserve(self.arb.get_x_tkn())
        self.y_amt_arb = self.lp.get_reserve(self.arb.get_y_tkn())
        self.price_arb = self.lp.get_price(self.arb.get_x_tkn())            
                 


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...utils.data import UniswapExchangeData
from ...cpt.factory import UniswapFactory
from ...simulate import SolveDeltas
from ...process.deposit import SwapDeposit
from ...process.swap import WithdrawSwap

USER_NM = 'USER_SIM'

class SimpleLPSimulation:
    
    def __init__(self):
        self.lp = None
        self.tkn_price_arr = []
        self.x_amt_arr = []
        self.y_amt_arr = []
        self.tkn_x_amt = None
        self.tkn_y_amt = None

    def init_amts(self, tkn_x_amt, init_price):
        self.tkn_x_amt = tkn_x_amt
        self.tkn_y_amt = init_price*tkn_x_amt 

    def create_lp(self, tkn_x, tkn_y):
        
        assert self.tkn_x_amt != None or self.tkn_y_amt != None, 'SimpleLPSimulation: TKN AMTS NOT INITIALIZED'
        
        factory = UniswapFactory("TKN pool factory", None)
        
        exchg_data = UniswapExchangeData(tkn0 = tkn_x, tkn1 = tkn_y, symbol='LP', address=None)
        self.lp = factory.deploy(exchg_data)
        self.lp.add_liquidity(USER_NM, self.tkn_x_amt, self.tkn_y_amt, self.tkn_x_amt, self.tkn_y_amt)

    def run(self, p_trial_arr):
        sDel = SolveDeltas(self.lp)
        tkn_x = self.lp.factory.token_from_exchange[self.lp.name][self.lp.token0]
        tkn_y = self.lp.factory.token_from_exchange[self.lp.name][self.lp.token1]
        tkn_price_arr = []
        lp_tot_arr = []
        x_amt_arr = []
        y_amt_arr = []
        for p in p_trial_arr[1:]: 
            
            swap_dx, swap_dy = sDel.calc(p)   
            if(swap_dx >= 0):
                expected_amount_dep = SwapDeposit().apply(self.lp, tkn_x, USER_NM, abs(swap_dx))
                expected_amount_out = WithdrawSwap().apply(self.lp, tkn_y, USER_NM, abs(swap_dy))
            elif(swap_dy >= 0):
                expected_amount_dep = SwapDeposit().apply(self.lp, tkn_y, USER_NM, abs(swap_dy))
                expected_amount_out = WithdrawSwap().apply(self.lp, tkn_x, USER_NM, abs(swap_dx)) 
              
            # ************************* #
            # do advanced lp stuff here
            # ************************* #
                
            self.tkn_price_arr.append(self.lp.get_price(tkn_x))    
            self.x_amt_arr.append(self.lp.get_reserve(tkn_x))  
            self.y_amt_arr.append(self.lp.get_reserve(tkn_y))  
    
    def get_lp(self):
        return self.lp
    
    def get_tkn_price_sim(self):
        return self.tkn_price_arr
    
    def get_x_amt_sim(self):
        return self.x_amt_arr
    
    def get_y_amt_sim(self): 
        return self.y_amt_arr
        


# ConstantIPS.py
# Author: Ian Moore ( utiliwire@gmail.com )
# Date: Sept 2022

from .IPS import IPS

class ConstantIPS(IPS):
       
    def __swap_balance(self, balance, price):
        
        if price == None:
            return balance
        else:
            return balance*price
        
    def calc_ips_from_state(self, state, price = None):
              
        A0 = state.get_balance() - state.get_yield() - state.get_delta()
        A0 = self.__swap_balance(A0, price)
        
        dt = state.get_event().get_time_delta()
        
        A1 = A0 + state.get_yield()
        A1 = self.__swap_balance(A1, price)
        
        return self.calc_ips(A0, A1, dt)
    
    def calc_ips(self, a0, a1, dt):
        return (a1/a0)**(1/dt) - 1
        


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..index import RebaseIndexToken
from ..index import SettlementLPToken
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import UniV3Helper

class LPQuote():
    
    """ 
        Liquidity pool token quotes (ie, price, reserve and liquidity)
        
        Parameters
        -----------------
        quote_opposing : boolean
            Quote the opposing token amount by default, given LP and a token    
    """      
    
    def __init__(self, quote_opposing = True, include_fee = False):
        self.quote_opposing = quote_opposing
        self.include_fee = include_fee
        self.price_tkn = True
        
    def get_opposing_token(self, lp, tkn):
        
        """ get_x

            Get opposing token given LP exchange and one of its token assetes
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair                  

            Returns
            -----------------
            tkn: ERC20
                Opposing token from CPT pair      
        """             
        
        opposing_tkn_nm = lp.token1 if(tkn.token_name == lp.token0) else lp.token0
        return lp.factory.token_from_exchange[lp.name][opposing_tkn_nm]

    def get_reserve(self, lp, token, lwr_tick = None, upr_tick = None):  
        
        """ get_reserve

            Get reserve amount of token given LP exchange 
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            token: ERC20
                Token asset from CPT pair                  

            Returns
            -----------------
            reserve: float
                Reserve amount from token asset     
        """             
        if(lp.version == UniswapExchangeData.VERSION_V2):
            if(token.token_name == lp.token0):        
                reserve_out = lp.convert_to_human(lp.reserve0) 
            elif(token.token_name == lp.token1):
                reserve_out = lp.convert_to_human(lp.reserve1)   

            if token.type == 'index':
                self.quote_opposing = False
                parent_lp = token.parent_lp
                parent_token = token.parent_tkn
                reserve_out = self.get_amount_from_lp(parent_lp, parent_token, reserve_out)  
        
        elif(lp.version == UniswapExchangeData.VERSION_V3):

            if(lp.total_supply == 0):
                reserve_out = 0
            else:    
                reserve_out = lp.get_virtual_reserve(token) 
                if token.type == 'index':
                    self.quote_opposing = False
                    parent_lp = token.parent_lp
                    parent_token = token.parent_tkn
                    parent_lp_tkn_x = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token0]
                    p = parent_lp.get_price(parent_lp_tkn_x)
                    lwr_tick = UniV3Helper().get_price_tick(parent_lp, -1, p, 1000)
                    upr_tick = UniV3Helper().get_price_tick(parent_lp, 1, p, 1000)
                    reserve_out = self.get_amount_from_lp(parent_lp, parent_token, reserve_out, lwr_tick, upr_tick)         
                    
        return reserve_out

    def get_price(self, lp, tkn, lwr_tick = None, upr_tick = None):  
        
        """ get_price

            Get price of token given LP exchange 
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair                  

            Returns
            -----------------
            price: float
                Token price with respect to opposing token    
        """           
           
        opposing_tkn = self.get_opposing_token(lp, tkn)

        if(tkn.token_name == lp.token0):
            reserve0 = self.get_reserve(lp, tkn, lwr_tick, upr_tick)
            reserve1 = self.get_reserve(lp, opposing_tkn, lwr_tick, upr_tick)
            if(reserve0 == 0):
                return 0 
            else:
                return reserve1/reserve0 if self.price_tkn else lp.get_price(tkn)
        elif(tkn.token_name == lp.token1):    
            reserve1 = self.get_reserve(lp, tkn, lwr_tick, upr_tick)
            reserve0 = self.get_reserve(lp, opposing_tkn, lwr_tick, upr_tick)        
            if(reserve1 == 0):
                return 0
            else:
                return reserve0/reserve1 if self.price_tkn else lp.get_price(tkn)
        else:
            print('ERROR: wrong input token')              
        
    def get_liquidity(self, lp, tkn, amount_in): 
        
        """ get_liquidity

            Get liquidity amount given amount of one of the tokens in x*y CPT pair 
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair   
            amount_in: float
                Amount of liquidity               

            Returns
            -----------------
            amt_out: float
                Liquidity amount given amount of one of the tokens in x*y CPT pair   
        """           
        
        if(tkn.token_name == lp.token0):
            amt_out = (amount_in * lp.total_supply)/lp.reserve0 
        else:
            amt_out = (amount_in * lp.total_supply)/lp.reserve1 

        return amt_out         
        
        
    def get_amount(self, lp, tkn, amount_in, lwr_tick = None, upr_tick = None):
        
        """ get_amount

            Get amount of reserve for opposing token, given an amount of reserve of input token
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair   
            amount_in: float
                Amount of reserve token               

            Returns
            -----------------
            amt_out: float
                Amount of reserve for opposing token 
        """       

        if (amount_in == 0):
            return 0

        if(lp.version == UniswapExchangeData.VERSION_V2):

            if(self.include_fee):
                if(tkn.token_name == lp.token0):
                    amt_out = lp.get_amount_out0(amount_in)
                else:
                    amt_out = lp.get_amount_out1(amount_in)
            else:
                if(tkn.token_name == lp.token0):
                    amt_out = (amount_in * lp.reserve1) / lp.reserve0
                else:
                    amt_out = (amount_in * lp.reserve0) / lp.reserve1

        elif(lp.version == UniswapExchangeData.VERSION_V3): 
            quote_out = UniV3Helper().quote(lp, tkn, amount_in, lwr_tick, upr_tick)
            amt_out = quote_out[0]
    
        return amt_out if self.quote_opposing else amount_in 
    
    def get_amount_from_lp(self, lp, tkn, amount_lp_in, lwr_tick = None, upr_tick = None):
        
        """ get_amount_from_lp

            Get amount of opposing token, given an amount priced in liquidity
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair   
            amount_lp_in: float
                Amount of liquidity               

            Returns
            -----------------
            amt_out: float
                Amount of opposing token 
        """           
        
        if(amount_lp_in > 0):
            itkn_amt = RebaseIndexToken().apply(lp, tkn, amount_lp_in, lwr_tick, upr_tick)
            amt_out = self.get_amount(lp, tkn, itkn_amt, lwr_tick, upr_tick) if self.quote_opposing else itkn_amt
        else:
            amt_out = 0
            
        if(lp.version == UniswapExchangeData.VERSION_V2):   
            return lp.convert_to_human(amt_out) 
        elif(lp.version == UniswapExchangeData.VERSION_V3): 
            return amt_out  

    
    
    def get_lp_from_amount(self, lp, tkn, amount_in, lwr_tick = None, upr_tick = None):
        
        """ get_lp_from_amount

            Get amount of liquidity, given an amount reserve token
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair   
            amount_lp_in: float
                Amount of reserve token               

            Returns
            -----------------
            amt_out: float
                Amount of liquidity 
        """         
        
        if(amount_in > 0):
            lp_amt = SettlementLPToken().apply(lp, tkn, amount_in, lwr_tick, upr_tick)
        else:
            lp_amt = 0

        if(lp.version == UniswapExchangeData.VERSION_V2):   
            return lp.convert_to_human(lp_amt) 
        elif(lp.version == UniswapExchangeData.VERSION_V3): 
            return lp_amt  
         


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from .ERC20 import ERC20

class LPERC20(ERC20):
    
    """ DAOSYS ERC20 token

        Parameters
        -----------------
        token_name : str
            Token name 
        token_addr : str
            Token address  
        token_total : float
            Token holdings 
        type : float
            Token type 
    """      

    def __init__(self, name: str, addr: str) -> None:
        self.token_name = name+'-LP'
        self.token_addr = addr
        self.token_supply = 1_000_000_000
        self.token_total = 0
        self.type = 'lp'
        
    def set_token_lp(self, token_total):
        
        """ set_token_lp

            Reset token LP
                
            Parameters
            -----------------
            token_total : float
                token total        
        """         
        
        self.token_total = token_total

    def add_token_lp(self, value):
        
        """ add_token_lp

            Add token delta to token total
                
            Parameters
            -----------------
            value : float
                token delta        
        """            
        
        self.token_total += value           
        
    def remove_token_lp(self, value):
        
        """ remove_token_lp

            Remove token delta to token total
                
            Parameters
            -----------------
            value : float
                token delta        
        """          
        
        self.token_total -= value        




# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np

class TimeDeltaModel():   
    
    
    def __init__(self, no_time_delay = False):
        self.__no_time_delay = no_time_delay

    def apply(self, n = 1, p = 0.00001):  
        
        if(self.__no_time_delay):
            return [0] * n if n > 1 else 0
        elif(n == 1):
            return np.random.negative_binomial(1, p)  
        else:
            
            res = []
            for k in range(n):
                rval = np.random.negative_binomial(1, p)
                res.append(rval)
                
            return np.array(res)              

            
 


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import string
import random

class IDGenerator():

    def get_chars(self):
        return string.ascii_lowercase

    def get_digits(self):
        return string.digits     
    
    def apply(self, n=6):
        chars =  self.get_chars() + self.get_digits() 
        return ''.join(random.choice(chars) for _ in range(n))
        
   


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...math.basic import IDGenerator

ADDRESS_PREFIX = '0x'
ADDRESS_LENGTH = 40

class MockAddress():

    def __init__(self):
        pass

    def apply(self, n = 1, addr_len = None):
        if(n == 1):
            return self.gen_address(addr_len)
        else:
            out = []
            for k in range(n):
                out.append(self.gen_address(addr_len))
            return out
    
    def gen_address(self, addr_len): 
        addr_len = ADDRESS_LENGTH if addr_len == None else addr_len
        return ADDRESS_PREFIX+IDGenerator().apply(addr_len)


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Original MIT licenced Position file from chainflip-io 
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

from . import LiquidityMath, FullMath

from .Shared import *

### @title Position
### @notice Positions represent an owner address' liquidity between a lower and upper tick boundary
### @dev Positions store additional state for tracking fees owed to the position.
@dataclass
class PositionInfo:
    ## the amount of liquidity owned by this position
    liquidity: int
    ## fee growth per unit of liquidity as of the last update to liquidity or fees owed
    feeGrowthInside0LastX128: int
    feeGrowthInside1LastX128: int
    ## the fees owed to the position owner in token0#token1 => uint128
    tokensOwed0: int
    tokensOwed1: int


### @notice Returns the Info struct of a position, given an owner and position boundaries
### @param self The mapping containing all user positions
### @param owner The address of the position owner
### @param tickLower The lower tick boundary of the position
### @param tickUpper The upper tick boundary of the position
### @return position The position info struct of the given owners' position
def get(self, owner, tickLower, tickUpper):
    checkInputTypes(account=owner, int24=(tickLower, tickUpper))

    # Need to handle non-existing positions in Python
    key = hash((owner, tickLower, tickUpper))
    if not self.__contains__(key):
        # We don't want to create a new position if it doesn't exist!
        # In the case of collect we add an assert after that so it reverts.
        # For mint there is an amount > 0 check so it is OK to initialize
        # In burn if the position is not initialized, when calling Position.update it will revert with "NP"
        self[key] = PositionInfo(0, 0, 0, 0, 0)
    return self[key]


def assertPositionExists(self, owner, tickLower, tickUpper):
    checkInputTypes(account=owner, int24=(tickLower, tickLower))
    positionInfo = get(self, owner, tickLower, tickUpper)
    assert positionInfo != PositionInfo(0, 0, 0, 0, 0), "Position doesn't exist"
    return positionInfo


### @notice Credits accumulated fees to a user's position
### @param self The individual position to update
### @param liquidityDelta The change in pool liquidity as a result of the position update
### @param feeGrowthInside0X128 The all-time fee growth in token0, per unit of liquidity, inside the position's tick boundaries
### @param feeGrowthInside1X128 The all-time fee growth in token1, per unit of liquidity, inside the position's tick boundaries
def update(self, liquidityDelta, feeGrowthInside0X128, feeGrowthInside1X128):
    checkInputTypes(
        int128=(liquidityDelta), uint256=(feeGrowthInside0X128, feeGrowthInside1X128)
    )

    if liquidityDelta == 0:
        # Removed because a check is added for burn 0 uninitialized position
        # assert self.liquidity > 0, "NP"  ## disallow pokes for 0 liquidity positions
        liquidityNext = self.liquidity
    else:
        liquidityNext = LiquidityMath.addDelta(self.liquidity, liquidityDelta)

    ## calculate accumulated fees. Add toUint256 because there can be an underflow
    tokensOwed0 = FullMath.mulDiv(
        toUint256(feeGrowthInside0X128 - self.feeGrowthInside0LastX128),
        self.liquidity,
        FixedPoint128_Q128,
    )
    tokensOwed1 = FullMath.mulDiv(
        toUint256(feeGrowthInside1X128 - self.feeGrowthInside1LastX128),
        self.liquidity,
        FixedPoint128_Q128,
    )

    # NOTE: TokensOwed can be > MAX_UINT128 and < MAX_UINT256. Uniswap cast tokensOwed into uint128. This in itself
    # is an overflow and it can overflow again when adding self.tokensOwed0 += tokensOwed0. Uniswap finds this
    # acceptable to save gas and that behaviour is reproduced here.

    # Mimic Uniswap's solidity code overflow - uint128(tokensOwed0)
    if tokensOwed0 > MAX_UINT128:
        tokensOwed0 = tokensOwed0 & (2**128 - 1)
    if tokensOwed1 > MAX_UINT128:
        tokensOwed1 = tokensOwed1 & (2**128 - 1)

    ## update the position
    if liquidityDelta != 0:
        self.liquidity = liquidityNext
    self.feeGrowthInside0LastX128 = feeGrowthInside0X128
    self.feeGrowthInside1LastX128 = feeGrowthInside1X128

    if tokensOwed0 > 0 or tokensOwed1 > 0:
        # NOTE: For now we allow overflow to happen because in uniswap overflow is acceptable,
        # LPs has to withdraw before you hit type(uint128).max fees
        self.tokensOwed0 += tokensOwed0
        self.tokensOwed1 += tokensOwed1



from ..quote import LPQuote
from ...utils.tools.v3 import UniV3Helper

class ChildLP():
    
    def __init__(self, lp):
        self.lp = lp 
        self.version = lp.version
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        self.x_tkn = tokens[lp.token0]
        self.y_tkn = tokens[lp.token1]  
        self.token0 = self.x_tkn.token_name
        self.token1 = self.y_tkn.token_name
        self.name = lp.name 
        self.factory = lp.factory
        self.tick_space = 1000
        self.reserve0 = None
        self.reserve1 = None
        self.virtual_reserve0 = None
        self.virtual_reserve1 = None   

    def summary(self):
        self._update()          
        print(f"Exchange {self.lp.name} ({self.lp.symbol})") 
        print(f"Real Reserves:   {self.x_tkn.token_name} = {self.reserve0}, {self.y_tkn.token_name} = {self.reserve1}")
        print(f"Virtual Reserves:   {self.x_tkn.token_name} = {self.virtual_reserve0}, {self.y_tkn.token_name} = {self.virtual_reserve1}")
        print(f"Price ({self.y_tkn.token_name}/{self.x_tkn.token_name}): {self.virtual_reserve1/self.virtual_reserve0}\n")
      
    def _update(self):
        self.reserve0, self.virtual_reserve0 = self._get_reserves(self.x_tkn)
        self.reserve1, self.virtual_reserve1 = self._get_reserves(self.y_tkn)

    def _get_reserves(self, tkn):
        reserve = self.lp.get_reserve(tkn)
        if(reserve == 0):
            virtual_reserve = 0
        else:      
            virtual_reserve = self.lp.get_virtual_reserve(tkn)
            if tkn.type == 'index':
                parent_lp = tkn.parent_lp
                parent_token = tkn.parent_tkn
                parent_lp_tkn_x = parent_lp.factory.token_from_exchange[parent_lp.name][parent_lp.token0]
                p = parent_lp.get_price(parent_lp_tkn_x)
                lwr_tick = UniV3Helper().get_price_tick(parent_lp, -1, p, 1000)
                upr_tick = UniV3Helper().get_price_tick(parent_lp, 1, p, 1000)
                reserve = LPQuote(False).get_amount_from_lp(parent_lp, parent_token, int(reserve), lwr_tick, upr_tick)
                virtual_reserve = LPQuote(False).get_amount_from_lp(parent_lp, parent_token, int(virtual_reserve), lwr_tick, upr_tick)
            
        return reserve, virtual_reserve

    def get_price(self, tkn):  
        self._update()  
        if(tkn.token_name == self.x_tkn.token_name):
            return self.virtual_reserve1/self.virtual_reserve0 
        elif(tkn.token_name == self.y_tkn.token_name):
            return self.virtual_reserve0/self.virtual_reserve1  
    
    def get_reserve(self, tkn):
        self._update()
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        x_tkn = tokens[self.lp.token0]
        y_tkn = tokens[self.lp.token1]        
        
        if(tkn.token_name == self.x_tkn.token_name):
            return self.reserve0 
        elif(tkn.token_name == self.y_tkn.token_name):
            return self.reserve1 

    def get_base_tkn(self, tkn):
        if tkn.type == 'index':
            base_lp = tkn.parent_lp
            base_token = tkn.parent_tkn
            return base_token
        else:
            return tkn

    def get_real_reserve(self, tkn):  
        return self.lp.get_reserve(tkn)

    def get_virtual_reserve(self, tkn): 
        self._update() 
        if(tkn.token_name == self.x_tkn.token_name):
            return self.virtual_reserve0  
        elif(tkn.token_name == self.y_tkn.token_name):
            return self.virtual_reserve1         
        


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import requests

from ..data import Chain0x

class API0x():
    
    """ 0x API class which calls Chain0x data class for data
    
        Parameters
        -----------------
        chain : str
            API endpoint url indicating which chain to make API call (eg, api.0x.org)
        api_key : str
            0x API key (see dashboard.0x.org for more info)  
    
    """      

    def __init__(self, chain: str = None, api_key: str = None) -> None:
        self.chain = Chain0x.ETHEREUM if chain == None else chain
        self.api_key = Chain0x().get_api_key() if api_key == None else api_key
        
    def apply(self, sell_tkn: str = None, buy_tkn: str = None, sell_amt: str = None):
        
        """ apply

            Apply call to 0x API 
            
            Parameters
            -----------------
            buy_token : str
                Buy token contract address (ie, pulled from Chain0x data class)
            sell_token : str
                Buy token contract address (ie, pulled from Chain0x data class)  
            sell_amt : str
                0x API requirement (set to 10000000 by default)                
                
            Returns
            -----------------
            0x api data : dictionary
                JSON structured 0x api call data                 
        """         
        
        sell_tkn = Chain0x().get_sell_token() if sell_tkn == None else sell_tkn
        buy_tkn = Chain0x().get_buy_token() if buy_tkn == None else buy_tkn     
        sell_amt = Chain0x().get_api_sell_amount() if sell_amt == None else sell_amt 
        
        url = f'https://{self.chain}/swap/v1/quote?sellToken={sell_tkn}&buyToken={buy_tkn}&sellAmount={sell_amt}'
        
        headers = {
                'Content-type': 'application/json',
                '0x-api-key': self.api_key
            }

        search_response = requests.get(url, headers=headers).json()
        return search_response if 'price' in search_response else {}
        


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from abc import *

class Vault(ABC):
             
    @abstractmethod        
    def rebase_index_tkn(self, lp_token, token):
        pass
    @abstractmethod     
    def deposit_lp_tkn(self, _to, token, amt):
        pass    
    @abstractmethod     
    def remove_lp_tkn(self, _to, token, amt):
        pass
    @abstractmethod 
    def get_tkn_pair_amount(self, lp_tkn, tkn, liq):
        pass     
    @abstractmethod 
    def get_token_type(self, lp_tkn, tkn, liq):
        pass      


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np
import pandas as pd
from ...cpt.quote import LPQuote
from ...process.swap import Swap
from ...math.model import TokenDeltaModel
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import UniV3Helper
from ...utils.tools.v3 import UniV3Utils

class Arbitrage():
    
    THRES = 0.2
    FRAC = 0.1
    
    def __init__(self, lp, mstate, tDel = None):   
        self.lp = lp
        self.mstate = mstate
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
        self.threshold = self.THRES
        self.net_y = 0
        self.net_x = 0
        self.y_tot = 0
        self.x_tot = 0 
        self.lwr_tick = None
        self.upr_tick = None         
           
    def apply(self, price_benchmark, user_nm, amt_in = None):
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        
        x_tkn = tokens[self.lp.token0]
        y_tkn = tokens[self.lp.token1]
        
        amt_x_sell = 0; amt_y_buy = 0 
        amt_y_sell = 0; amt_x_buy = 0
        
        lwr_tick, upr_tick = self.gen_ticks(x_tkn)
        p_x = LPQuote().get_price(self.lp, x_tkn, lwr_tick, upr_tick)
        num_states = len(self.mstate.states)
        held_y_amt = self.mstate.get_current_state('dHeld') 
        
        # arbitrage LP1
        if(p_x != None and p_x > price_benchmark):
            while(p_x > price_benchmark):
                lwr_tick, upr_tick = self.gen_ticks(x_tkn)
                amt = self.gen_random_amt() if amt_in == None else self.FRAC*amt_in
                amt_x_sell += amt
                amt_y_buy += Swap().apply(self.lp, x_tkn, user_nm, amt)
                p_x = LPQuote().get_price(self.lp, x_tkn, lwr_tick, upr_tick)
                #print('[1] p-bench {:.3f} p_x {:.3f}'.format(price_benchmark, p_x))

        elif(p_x != None and p_x <= price_benchmark and num_states > 3):
            while(p_x <= price_benchmark):  
                lwr_tick, upr_tick = self.gen_ticks(x_tkn)
                amt = self.gen_random_amt() if amt_in == None else self.FRAC*amt_in
                q_amt = LPQuote().get_amount(self.lp, x_tkn, amt, lwr_tick, upr_tick)
                #q_held_x_amt = LPQuote().get_amount(self.lp, x_tkn, held_x_amt, lwr_tick, upr_tick)
                if(q_amt+amt_y_sell > self.threshold*held_y_amt): 
                #if(q_amt+amt_y_sell > self.threshold*held_x_amt):     
                    break
                else:  
                    amt_y_sell += q_amt
                    amt_x_buy += Swap().apply(self.lp, y_tkn, user_nm, q_amt)
                    p_x = LPQuote().get_price(self.lp, x_tkn, lwr_tick, upr_tick)
                #print('[2] p-bench {:.3f} p_x {:.3f}'.format(price_benchmark, p_x))    
        
        self.net_x = amt_x_buy - amt_x_sell
        self.net_y = amt_y_buy - amt_y_sell 
        
        self.y_tot += abs(amt_y_buy) + abs(amt_y_sell) 
        self.x_tot += abs(amt_x_buy) + abs(amt_x_sell)

    def set_portion_threshold(self, thres):            
        self.threshold = thres  

    def set_ticks(self, lwr_tick, upr_tick):
        self.lwr_tick = lwr_tick
        self.upr_tick = upr_tick
    
    def gen_ticks(self, x_tkn):
        if(self.lp.version == UniswapExchangeData.VERSION_V3):    
            fee = UniV3Utils.FeeAmount.MEDIUM
            tick_spacing = UniV3Utils.TICK_SPACINGS[fee]
            lwr_tick = UniV3Utils.getMinTick(tick_spacing)
            upr_tick = UniV3Utils.getMaxTick(tick_spacing)
        else:
            lwr_tick = None
            upr_tick = None 
        return lwr_tick, upr_tick
        
    def gen_random_amt(self):            
        return self.FRAC*self.tDel.delta()        
                   
    def update_state(self, token):
        net_amt = self.net_amount(token)
        self.mstate.update_current_state(net_amt) 
    
    def get_x_tot(self):
        return self.x_tot 
    
    def get_y_tot(self):
        return self.y_tot     
    
    def net_amount(self, token):
        tokens = self.lp.factory.token_from_exchange[self.lp.name]
        x_tkn = tokens[self.lp.token0]
        y_tkn = tokens[self.lp.token1]
        net_amt = 0
        
        if(token.token_name == x_tkn.token_name):
            net_amt = self.net_x
        elif(token.token_name == y_tkn.token_name):
            net_amt = self.net_y
            
        return net_amt    


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Modified version of original MIT licenced TickMath file from chainflip-io
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

from . import TickMath

### @title Math library for liquidity

### @notice Add a signed liquidity delta to liquidity and revert if it overflows or underflows
### @param x The liquidity before change
### @param y The delta by which liquidity should be changed
### @return z The liquidity delta
def addDelta(x, y):
    if y < 0:
        z = x - abs(y)
        # Mimic solidity underflow
        assert z >= 0, "LS"
    else:
        z = x + abs(y)
        # Mimic solidity overflow check
        assert z <= TickMath.MAX_UINT128, "LA"
    return z



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from scipy.optimize import fsolve
from ...utils.data import UniswapExchangeData
import warnings
warnings.filterwarnings("ignore")

class SolveDeltas():
    
    def __init__(self, lp):
        self.lp = lp
        self.tkn_x = lp.factory.token_from_exchange[lp.name][lp.token0]
        self.tkn_y = lp.factory.token_from_exchange[lp.name][lp.token1]      
        self.x = self._get_reserve(self.tkn_x) 
        self.y = self._get_reserve(self.tkn_y)       
        self.p = lp.get_price(self.tkn_x)
        self.p_prev = lp.get_price(self.tkn_x)
        self.dp = 0
              
    def get_lp(self):
        return self.lp   
                    
    def func_swap_yx(self, z):
        z[0] = 1 if z[0] == 0 else z[0]
        return [(self.x*abs(z[1]) + abs(z[0])*self.y)/(self.x**2 - abs(z[0])*self.x) - self.dp,
                abs(z[1])/abs(z[0]) - self.p]

    def func_swap_xy(self, z):
        z[0] = 1 if z[0] == 0 else z[0]
        return [-(self.x*abs(z[1]) + abs(z[0])*self.y)/(self.x**2 + abs(z[0])*self.x) - self.dp,
                abs(z[1])/abs(z[0]) - self.p]
        
    def calc(self, p, x0 = None, fac = None):
        self.p = p
        self.p_prev = self.lp.get_price(self.tkn_x) 
        self.dp = p - self.p_prev
        self.x = self._get_reserve(self.tkn_x)
        self.y = self._get_reserve(self.tkn_y)   
        fac = 0.1 if fac == None else fac
        dx0 = 0.5 if x0 == None else x0/p
        dy0 = 0.5 if x0 == None else x0
        if(self.dp >= 0):
            dx, dy = fsolve(self.func_swap_yx, [dx0, dy0], xtol=1e-6, factor=fac, maxfev = 200)
            self.p_prev = p
            return -dx, dy
        else:
            dx, dy = fsolve(self.func_swap_xy, [dx0, dy0], xtol=1e-6, factor=fac, maxfev = 200)
            self.p_prev = p
            return dx, -dy  

    def _get_reserve(self, tkn):
        
        if(self.lp.version == UniswapExchangeData.VERSION_V2):
            return self.lp.get_reserve(tkn) 
        elif(self.lp.version == UniswapExchangeData.VERSION_V3):
            return self.lp.get_virtual_reserve(tkn)


from ..exchg.ChildLP import ChildLP
from ...utils.data import UniswapExchangeData

class QuoteLiquidity():

    def __init__(self):
        pass

    def apply(self, lp_in, tkn):
        tkn_x = lp_in.factory.token_from_exchange[lp_in.name][lp_in.token0]
        tkn_y = lp_in.factory.token_from_exchange[lp_in.name][lp_in.token1]
        
        return self.get_tkn_amt(lp_in, tkn, tkn_x) + self.get_tkn_amt(lp_in, tkn, tkn_y)
    
    def get_base_lp(self, lp_in):
    
        tkn_x = lp_in.factory.token_from_exchange[lp_in.name][lp_in.token0]
        tkn_y = lp_in.factory.token_from_exchange[lp_in.name][lp_in.token1]
        
        if(tkn_x.type == 'index'):
            base_lp = tkn_y.parent_lp
        elif(tkn_y.type == 'index'):
            base_lp = tkn_y.parent_lp
        else: 
            base_lp = lp_in
            
        return base_lp  
    
    def get_tkn_amt(self, lp_in, tkn_in, tkn):

        if(lp_in.version == UniswapExchangeData.VERSION_V2):    

            if tkn.type == 'index':
                base_lp = tkn.parent_lp
                base_tkn = tkn.parent_tkn
                base_tkn_amt = base_lp.get_reserve(base_tkn)
            else:
                base_tkn = tkn
                base_tkn_amt = lp_in.get_reserve(tkn)

        elif(lp_in.version == UniswapExchangeData.VERSION_V3):
            lp_child = ChildLP(lp_in)
            base_tkn = lp_child.get_base_tkn(tkn)
            base_tkn_amt = lp_child.get_reserve(tkn)
            
        base_lp = self.get_base_lp(lp_in)
        
        tot_amt = 0
        if(tkn_in.token_name == base_tkn.token_name):
            tot_amt += base_tkn_amt
        else:
            tot_amt += base_tkn_amt*base_lp.get_price(base_tkn)
    
        return tot_amt


# AggregateIPS.py
# Author: Ian Moore ( utiliwire@gmail.com )
# Date: Sept 2022

from ..IPS import IPS

class AggregateIPS():
    
    def __init__(self, ips, states = []):
        self.__states = states
        self.__ips = ips
        
    def update_states(self, states):
        self.__states = states

    def apply(self, states):
        aggr = 0
        for k in range(len(states)):
            aggr += self.__ips.calc_ips_from_state(states[k])
        return aggr/len(states)    
        


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..exchg import UniswapExchange 
from ..exchg import UniswapV3Exchange 
from ...erc import ERC20
from ...erc import LPERC20
from ...utils.interfaces import IExchangeFactory 
from ...utils.data import UniswapExchangeData
from ...utils.data import FactoryData

class UniswapFactory(IExchangeFactory):

    """ 
        Create Uniswap liquidity pools for given token pairs
        
        Parameters
        -----------------
        name : str
            Token name 
        address : str
            Token 0 name  
        exchange_from_token : dictionary
            Map of tokens to exchanges
        tokens_from_exchange : dictionary
            Map of exchanges to pair tokens          
    """     
    
    def __init__(self, name: str, address: str) -> None:
        self.name = name
        self.address = address
        self.exchange_from_token = {}
        self.token_from_exchange = {}
        self.parent_lp = None

    def deploy(self, exchg_data : UniswapExchangeData):
        
        """ deploy

            Deploy a Uniswap liquidity pool (LP) exchange
                
            Parameters
            -----------------
            exchg_data : UniswapExchangeData
                Uniswap exchange initialization data     

            Returns
            -----------------
            exchange : UniswapExchange
                Newly created exchange that is also a LP token                    
        """  
        
        token0 = exchg_data.tkn0
        token1 = exchg_data.tkn1
        symbol = exchg_data.symbol
        address = exchg_data.address
        precision = exchg_data.precision

        assert symbol not in self.token_from_exchange, 'UniswapV2Factory: EXCHANGE_CREATED'
            
        self.parent_lp = token0.parent_lp if token0.type == 'index' else self.parent_lp
        self.parent_lp = token1.parent_lp if token1.type == 'index' else self.parent_lp 
        
        factory_struct = FactoryData(self.token_from_exchange,  self.parent_lp, self.name, self.address)

        match exchg_data.version:
            case UniswapExchangeData.VERSION_V2:
                exchg_struct = UniswapExchangeData(tkn0 = token0, tkn1 = token1, symbol=symbol, precision = precision, address=address)
                exchange = UniswapExchange(factory_struct, exchg_struct) 
            case UniswapExchangeData.VERSION_V3: 
                exchg_struct = UniswapExchangeData(tkn0 = token0, tkn1 = token1, symbol=symbol, 
                                                   address=address, version = UniswapExchangeData.VERSION_V3, 
                                                   precision = precision, 
                                                   tick_spacing = exchg_data.tick_spacing, fee = exchg_data.fee)                
                exchange = UniswapV3Exchange(factory_struct, exchg_struct) 
        
        self.exchange_from_token[token0.token_name] = exchange
        self.token_from_exchange[exchange.name] = {token0.token_name: token0, token1.token_name: token1}
        
        return exchange

    def get_exchange(self, token):
        
        """ get_exchange

            Get exchange from given pair token
                
            Parameters
            -----------------
            token : ERC20
                receiving user address      
                
            Returns
            -----------------
            exchange : UniswapExchange
                exchange from mapped token                    
        """         
        
        return self.token_from_exchange.get(token)

    def get_token(self, exchange):
        
        """ get_token

            Get token from exchange
                
            Parameters
            -----------------
            exchange : UniswapExchange
                receiving user address      
                
            Returns
            -----------------
            token : ERC20 
                token from mapped exchange                     
        """          
        
        return self.token_from_exchange.get(exchange)


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ...erc import LPERC20
from ...utils.interfaces import IExchange 
from ...utils.data import UniswapExchangeData
from ...utils.data import FactoryData
from ...utils.tools.v3 import UniV3Helper
from ...utils.tools import SaferMath
import math


#MINIMUM_LIQUIDITY = 1e-15
MINIMUM_LIQUIDITY = 1000

class UniswapExchange(IExchange, LPERC20):
    
    """ 
        Uniswap V2 Exchange  

        Parameters
        -----------------
        factory_struct : FactoryInit
            Factory initialization data
        exchg_struct : UniswapExchangeInit
            Exchange initialization data           
    """          
    def __init__(self, factory_struct: FactoryData, exchg_struct: UniswapExchangeData):
        super().__init__(exchg_struct.tkn0.token_name+exchg_struct.tkn1.token_name, exchg_struct.address)
        self.version = exchg_struct.version
        self.precision = exchg_struct.precision
        self.factory = factory_struct
        self.token0 = exchg_struct.tkn0.token_name     
        self.token1 = exchg_struct.tkn1.token_name       
        self.reserve0 = 0             
        self.reserve1 = 0       
        self.fee0_arr = []
        self.fee1_arr = []
        self.aggr_fee0 = 0
        self.aggr_fee1 = 0
        self.collected_fee0 = 0
        self.collected_fee1 = 0              
        self.name =  f"{self.token0}-{self.token1}"
        self.symbol = exchg_struct.symbol
        self.liquidity_providers = {}
        self.last_liquidity_deposit = 0
        self.total_supply = 0

    def summary(self):

        """ summary
            Summary print-out of exchange, reserves and liquidity               
        """  

        reserve0 = self.convert_to_human(self.reserve0)
        reserve1 = self.convert_to_human(self.reserve1)
        total_supply = self.convert_to_human(self.total_supply)
        
        print(f"Exchange {self.name} ({self.symbol})")
        print(f"Reserves: {self.token0} = {reserve0}, {self.token1} = {reserve1}")
        print(f"Liquidity: {total_supply} \n")    

    def add_liquidity(self, _from_addr, amountADesired, amountBDesired, amountAMin, amountBMin):
        
        """ add_liquidity

            Add liquidity to both tokens in the pair
                
            Parameters
            -----------------
            _from_addr : str
                receiving user address      
            amountADesired : float
                desired amount of A      
            amountBDesired : float
                desired amount of B   
            amountAMin : float
                Minimum amount of A  
            amountBMin : float
                Minimum amount of B                 
        """  

        amountADesired = self.convert_to_machine(amountADesired)
        amountBDesired = self.convert_to_machine(amountBDesired)
        amountAMin = self.convert_to_machine(amountAMin)
        amountBMin = self.convert_to_machine(amountBMin)

        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV2: TOKEN_UNAVAILABLE' 
        amountA, amountB = self._add_liquidity(amountADesired, amountBDesired, amountAMin, amountBMin)
        
        tokens.get(self.token0).deposit(_from_addr, int(amountA))
        tokens.get(self.token1).deposit(_from_addr, int(amountB))

        self.mint(_from_addr, amountA, amountB)
        return amountA, amountB

    def _add_liquidity(self, amountADesired, amountBDesired, amountAMin, amountBMin):
        
        """ _add_liquidity

            Add liquidity to both tokens in the pair
                
            Parameters
            -----------------    
            amountADesired : float
                desired amount of A      
            amountBDesired : float
                desired amount of B   
            amountAMin : float
                Minimum amount of A  
            amountBMin : float
                Minimum amount of B   
                
            Returns
            -------
            amount0 : float
                removed liquidity from reserve0    
            amount1 : float
                removed liquidity from reserve1                   
        """            
        
        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV2: TOKEN_UNAVAILABLE' 

        if self.reserve0 == 0 and self.reserve1 == 0:
            amountA = amountADesired
            amountB = amountBDesired
        else:
            amountBOptimal = self.quote(amountADesired, self.reserve0, self.reserve1)
            if amountBOptimal <= amountBDesired:                    
                assert round(int(amountBOptimal),-25) >= round(int(amountBMin),-25), 'UniswapV2: INSUFFICIENT_B_AMOUNT'
                amountA = amountADesired
                amountB = amountBOptimal
            else:
                amountAOptimal = self.quote(amountBDesired, self.reserve1, self.reserve0)                
                assert round(int(amountAOptimal),-20) <= round(int(amountADesired),-20)
                assert round(int(amountAOptimal),-20) >= round(int(amountAMin),-20), 'UniswapV2: INSUFFICIENT_A_AMOUNT'
                amountA = amountAOptimal
                amountB = amountBDesired

        return amountA, amountB   

    def get_amounts(self, to_addr, liquidity):
        
        """ get_amounts

            Remove liquidity from both coins in the pair based on lp amount
                
            Parameters
            -----------------
            to_addr : str
                receiving user address  
            liquidity : float
                lp amount                  
                
            Returns
            -------
            amountA : float
                liquidity portion of reserve0    
            amountB : float
                liquidity portion of reserve1                    
        """         
        
        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV2: TOKEN_UNAVAILABLE' 

        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total
        total_liquidity = self.liquidity_providers[to_addr]
        if liquidity >= total_liquidity:
            liquidity = total_liquidity

        amountA = SaferMath().mul_div_round(liquidity, balanceA, self.total_supply)
        amountB = SaferMath().mul_div_round(liquidity, balanceB, self.total_supply)
        
        return amountA, amountB    
    
    def remove_liquidity(self, to_addr, liquidity, amountAMin, amountBMin):
        
        """ remove_liquidity

            Remove liquidity from both coins in the pair based on lp amount
                
            Parameters
            -----------------
            to_addr : str
                receiving user address  
            liquidity : float
                lp amount to removed                 
            amountAMin : float
                Minimum amount of A  
            amountBMin : float
                Minimum amount of B 
                
            Returns
            -------
            amountA : float
                removed liquidity from reserve0    
            amountB : float
                removed liquidity from reserve1                    
        """   
     
        liquidity = self.convert_to_machine(liquidity)
        amountAMin = self.convert_to_machine(amountAMin)
        amountBMin = self.convert_to_machine(amountBMin)

        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV2: TOKEN_UNAVAILABLE' 

        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total
        total_liquidity = self.liquidity_providers[to_addr]
        if liquidity >= total_liquidity:
            liquidity = total_liquidity

        amountA = SaferMath().mul_div_round(liquidity, balanceA, self.total_supply)
        amountB = SaferMath().mul_div_round(liquidity, balanceB, self.total_supply)

        assert (round(int(amountA),-20) >= round(int(amountAMin),-20)), 'AMOUNTA {} AMOUNT_A_MIN {}'.format(round(amountA,5), round(amountAMin,5))
        assert amountA > 0 and amountB > 0, 'UniswapV2: INSUFFICIENT_LIQUIDITY_BURNED'
        assert round(int(amountA),-20) >= round(int(amountAMin),-20), 'UniswapV2: INSUFFICIENT_A_AMOUNT'
        assert round(int(amountB),-20) >= round(int(amountBMin),-20), 'UniswapV2: INSUFFICIENT_B_AMOUNT'
        
        self.burn(to_addr, liquidity, amountA, amountB)
        return self.convert_to_human(amountA), self.convert_to_human(amountB)

    def swap_exact_tokens_for_tokens(self, amountIn, amountOutMin, token_in, to_addr):
        
        """ swap_exact_tokens_for_tokens

            Swap amt of token for min opposing token out
                
            Parameters
            -----------------
            amountIn : float
                swap amount in
            amountOutMin : float
                min swap amount in                
            token_in : ERC20
                Token to be swapped
            to_addr : str
               receiving user address  
                
            Returns
            -------
            amount_out_expected : float
                amount expected from the swap                    
        """   
        amount_out_expected = self.get_amount_out(amountIn, token_in)

        assert amount_out_expected >= amountOutMin, 'UniswapV2: INSUFFICIENT_OUTPUT_AMOUNT'

        amount_out_expected = self.convert_to_machine(amount_out_expected)
        amountIn = self.convert_to_machine(amountIn)
        amountOutMin = self.convert_to_machine(amountOutMin)        

        tokens = self.factory.token_from_exchange[self.name]
        if(token_in.token_name == self.token0):
            tokens.get(self.token0).deposit(to_addr, amountIn)
            self.swap(0, amount_out_expected, to_addr)
        elif(token_in.token_name == self.token1):
            tokens.get(self.token1).deposit(to_addr, amountIn)
            self.swap(amount_out_expected, 0, to_addr)

        return self.convert_to_human(amount_out_expected)

    def burn(self, to_addr, liquidity, amountA, amountB):
        
        """ burn

            Burn liquidity from both coins in the pair based on lp amount
                
            Parameters
            -----------------
            to_addr : str
               receiving user address  
            liquidity : float
                amount of liquidity to be burned                  
            amountA : float
                est. amount from reserve0 to be burned
            amountB : float
                est. amount from reserve1 to be burned               
        """  
        
        self._burn(to_addr, liquidity)

        tokens = self.factory.token_from_exchange[self.name]
        tokens.get(self.token0).transfer(to_addr, amountA)
        tokens.get(self.token1).transfer(to_addr, amountB)

        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total

        self._update(balanceA, balanceB)

    def _burn(self, to_addr, value):
        
        """ _burn

            Burn liquidity from both coins in the pair based on lp amount
                
            Parameters
            -----------------
            to : str
               receiving user address  
            value : float
                amount of liquidity to be burned                           
        """            
        
        available_liquidity = self.liquidity_providers.get(to_addr)
        self.liquidity_providers[to_addr] = available_liquidity - value
        self.total_supply -= value

    def mint(self, to_addr, _amountA, _amountB):
        
        """ mint

            Mint new liquidity based on amounts on each coin in the pair
                
            Parameters
            -----------------
            to_addr : str
                receiving user address      
            _amountA : float
                desired amount of A      
            _amountB : float
                desired amount of B                   
        """  
        
        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV2: TOKEN_UNAVAILABLE' 

        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total


        amountA = balanceA - self.reserve0
        amountB = balanceB - self.reserve1
        
        assert round(int(amountA),-10) == round(int(_amountA),-10)
        assert round(int(amountB),-10) == round(int(_amountB),-10) 

        if self.total_supply != 0:
            liquidity = min(
                SaferMath().mul_div_round(amountA, self.total_supply, self.reserve0),
                SaferMath().mul_div_round(amountB, self.total_supply, self.reserve1)
            )
        else:
            liquidity = math.isqrt(amountA * amountB) - MINIMUM_LIQUIDITY
            self._mint("0", MINIMUM_LIQUIDITY)
            
        assert liquidity > 0, 'UniswapV2: INSUFFICIENT_LIQUIDITY_MINTED'
        
        self._update(balanceA, balanceB)
        self._mint(to_addr, liquidity)

    def _update(self, balanceA, balanceB):
        
        """ _update

            Update reserve amounts for both coins in the pair
                
            Parameters
            -----------------   
            balanceA : float
                new reserve amount of A      
            balance1 : float
                new reserve amount of B                   
        """         
        
        self.reserve0 = int(balanceA)  
        self.reserve1 = int(balanceB)

    def _mint(self, to_addr, value):
        
        """ _mint

            Update reserve amounts for both coins in the pair
                
            Parameters
            -----------------   
            to_addr : str
                receiving user address       
            value : float
                amount of new liquidity                  
        """          
        
        if self.liquidity_providers.get(to_addr):
            self.liquidity_providers[to_addr] += value
        else:
            self.liquidity_providers[to_addr] = value

        self.last_liquidity_deposit = value     
        self.total_supply += value
        
    def _tally_fees(self, fee0, fee1):
        
        """ _tally_fees

            Tally fee from swap and record last collected fee
                
            Parameters
            -----------------   
            fee0 : float
                fee from reserve0      
            fee1 : float
                fee from reserve1                 
        """         
        self.fee0_arr.append(fee0)
        self.fee1_arr.append(fee1)
        self.collected_fee0 += fee0 
        self.collected_fee1 += fee1        
        self.aggr_fee0 += fee0 
        self.aggr_fee1 += fee1
        
    def swap(self, amountA_out, amountB_out, to_addr):
        
        """ swap

            Remove liquidity from both coins in the pair based on lp amount
                
            Parameters
            -----------------
            amountA_out : float
                swap amountA out
            amountB_out : float
                swap amountB out               
            to_addr : str
               receiving user address                   
        """      
        
        assert amountA_out > 0 or amountB_out > 0, 'UniswapV2: INSUFFICIENT_OUTPUT_AMOUNT'
        assert amountA_out < self.reserve0 and amountB_out < self.reserve1, 'UniswapV2: INSUFFICIENT_LIQUIDITY'

        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0).token_addr != to_addr, 'UniswapV2: INVALID_TO_ADDRESS'
        assert tokens.get(self.token1).token_addr != to_addr, 'UniswapV2: INVALID_TO_ADDRESS'

        tokens.get(self.token0).transfer(to_addr, amountA_out)
        tokens.get(self.token1).transfer(to_addr, amountB_out)    
        
        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total

        amountA_in = balanceA - (self.reserve0 - amountA_out) if balanceA > self.reserve0 - amountA_out else 0
        amountB_in = balanceB - (self.reserve1 - amountB_out) if balanceB > self.reserve1 - amountB_out else 0
        assert amountA_in > 0 or amountB_in > 0, 'UniswapV2: INSUFFICIENT_INPUT_AMOUNT'

        balanceA_adjusted = balanceA * 1000 - amountA_in * 3  
        balanceB_adjusted = balanceB * 1000 - amountB_in * 3  
 
        adj_digits = max(len(str(balanceA_adjusted * balanceB_adjusted))-11, 1)
        lside = round(math.ceil(balanceA_adjusted * balanceB_adjusted), -adj_digits)
        rside = round(math.ceil(self.reserve0 * self.reserve1 * 1000**2), -adj_digits)

        # ******** FIX ********
        #assert  lside  ==  rside , 'UniswapV2: K'
    
        self._update(balanceA, balanceB)
        self._tally_fees(SaferMath().mul_div_round(amountA_in, 3, 1000), SaferMath().mul_div_round(amountB_in, 3, 1000))             
 
    def quote(self, amountA, reserveA, reserveB):
        
        """ quote

            Given amount asset and reserves, return equivalent amount of other asset
                
            Parameters
            -----------------
            amountA : float
                amount of a given asset token
            reserveA : float
                total amount of asset A in LP              
            reserveB : float
                total amount of asset B in LP                    
        """  

        amountA = self.convert_to_machine(amountA)
        reserveA = self.convert_to_machine(reserveA)
        reserveB = self.convert_to_machine(reserveB)
        
        assert amountA > 0, 'UniswapV2Library: INSUFFICIENT_AMOUNT'
        assert reserveA > 0 and reserveB > 0, 'UniswapV2Library: INSUFFICIENT_LIQUIDITY'
        quote_out = SaferMath().mul_div_round(amountA, reserveB, reserveA)   
        
        return self.convert_to_human(quote_out)

    def get_amount_out(self, amount_in, token_in):

        """ get_amount_out

            Get maximum output of opposing asset given input amount of token
                
            Parameters
            -----------------
            amount_in : float
                input amount of an asset
            token_in : ERC20
                asset token  
                          
            Returns
            -------
            amount out : float
                amount of opposing asset                       
        """ 
        
        if(token_in.token_name == self.token0):    
            return self.get_amount_out0(amount_in) 
        elif(token_in.token_name == self.token1):
            return self.get_amount_out1(amount_in) 
        else:
            return -1    
    
    def get_amount_out0(self, amount_in):
        
        """ get_amount_out0

            Get maximum amount of other asset given input amount of an asset
                
            Parameters
            -----------------
            amount_in : float
                input amount of an asset
                          
            Returns
            -------
            amount_out out : float
                reserve1 * reserve0       
        """

        amount_in = self.convert_to_machine(amount_in)
        
        assert amount_in > 0, 'UniswapV2Library: INSUFFICIENT_INPUT_AMOUNT'
        assert self.reserve0 > 0 and self.reserve1 > 0, 'UniswapV2Library: INSUFFICIENT_LIQUIDITY'

        amount_in_with_fee = amount_in * 997  
        amount_out =  SaferMath().div_round(amount_in * 997  * self.reserve1, self.reserve0 * 1000 + amount_in_with_fee)

        return self.convert_to_human(amount_out) 
    
    def get_amount_out1(self, amount_in):
        
        """ get_amount_out1

            Get maximum amount of other asset given input amount of an asset
                
            Parameters
            -----------------
            amount_in : float
                input amount of an asset
                          
            Returns
            -------
            amount_out out : float
                reserve1 * reserve0       
        """        

        amount_in = self.convert_to_machine(amount_in)
        
        assert amount_in > 0, 'UniswapV2Library: INSUFFICIENT_INPUT_AMOUNT'
        assert self.reserve0 > 0 and self.reserve1 > 0, 'UniswapV2Library: INSUFFICIENT_LIQUIDITY'

        amount_in_with_fee = amount_in * 997    
        amount_out = SaferMath().div_round(amount_in_with_fee * self.reserve0, self.reserve1 * 1000 + amount_in_with_fee)

        return self.convert_to_human(amount_out)  

    def update_reserves(self, user_nm, amountA_update = None, amountB_update = None):
        
        """ update_reserves

            Update reserve assets of either or both assets in the pair
                
            Parameters
            -----------------
            user_nm : str
                acccount holder
            amountA_update : float
                update amount of asset A             
            amountB_update : float
                update amount of asset B      
        """   
        
        amountA_update = amountA_update if amountA_update != None else self.reserve0
        amountB_update = amountB_update if amountB_update != None else self.reserve1
        self.burn(user_nm, self.liquidity_providers[user_nm], self.reserve0, self.reserve1)
        tokens = self.factory.token_from_exchange[self.name]
        tokens.get(self.token0).deposit(user_nm, amountA_update)
        tokens.get(self.token1).deposit(user_nm, amountB_update)
        self.mint(user_nm, amountA_update, amountB_update)                  
                    
    def get_price(self, token):  
        
        """ get_price

            Get price of select token in the exchange pair
                
            Parameters
            -----------------
            token : ERC20
                ERC20 token                
        """          

        if(token.token_name == self.token0):
            if(self.reserve0 == 0):
                return None 
            else:
                return self.reserve1/self.reserve0 
        elif(token.token_name == self.token1):
            if(self.reserve1 == 0):
                return None
            else:
                return self.reserve0/self.reserve1
                #return SaferMath().div_round(self.reserve0, self.reserve1)
        else:
            assert False, 'UniswapV2: WRONG_INPUT_TOKEN'      

    def get_last_liquidity_deposit(self):  
        
        """ get_last_liquidity_deposit

            Get the last liquidity deposit that went into pool         
        """          

        return self.convert_to_human(self.last_liquidity_deposit)     


    def get_liquidity_from_provider(self, provider_account):  
        
        """ get_liquidity_from_provider

            Get liquidity from provider  
            
            Parameters
            -----------------
            provider_account : str
                Provider account name  
            
        """          

        return self.convert_to_human(self.liquidity_providers[provider_account])      
                

    
    def get_liquidity(self):  
        
        """ get_liquidity

            Get liquidity of exchange pool         
        """          

        return self.convert_to_human(self.total_supply)       
            
    def get_reserve(self, token):  
        
        """ get_reserve

            Get reserve amount of select token in the exchange pair
                
            Parameters
            -----------------
            token : ERC20
                ERC20 token                
        """         
        
        if(token.token_name == self.token0):
            return self.convert_to_human(self.reserve0)
        elif(token.token_name == self.token1):
            return self.convert_to_human(self.reserve1) 
        else:
            assert False, 'UniswapV2: WRONG_INPUT_TOKEN'  

    def get_fee(self, token):  
        
        """ get_fee

            Get last fee amount of select token in the exchange pair
                
            Parameters
            -----------------
            token : ERC20
                ERC20 token                
        """         
        
        if(token.token_name == self.token0):
            return self.convert_to_human(self.collected_fee0)
        elif(token.token_name == self.token1):
            return self.convert_to_human(self.collected_fee1) 
        else:
            assert False, 'UniswapV2: WRONG_INPUT_TOKEN'  

    def convert_to_human(self, val): 
        val = val if self.precision == UniswapExchangeData.TYPE_GWEI else UniV3Helper().gwei2dec(val)
        return val

    def convert_to_machine(self, val): 
        val = int(val) if self.precision == UniswapExchangeData.TYPE_GWEI else UniV3Helper().dec2gwei(val)
        return val   
    
        


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Modified version of original MIT licenced UniswapPool class from chainflip-io
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

import math
from decimal import Decimal
from dataclasses import dataclass
from ...erc import LPERC20
from ...utils.interfaces import IExchange
from ...utils.data import FactoryData
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3.Shared import *
from ...utils.tools.v3 import Position, Tick, SqrtPriceMath, LiquidityMath
from ...utils.tools.v3 import SwapMath, TickMath, SafeMath, FullMath, UniV3Utils
from ...utils.tools.v3 import UniV3Helper

MINIMUM_LIQUIDITY = 1e-15
GWEI_PRECISION = 18

@dataclass
class Slot0:
    ## the current price
    sqrtPriceX96: int
    ## the current tick
    tick: int
    ## the current protocol fee as a percentage of the swap fee taken on withdrawal
    ## represented as an integer denominator (1#x)%
    feeProtocol: int
    
@dataclass
class ModifyPositionParams:
    ## the address that owns the position
    owner: str
    ## the lower and upper tick of the position
    tickLower: int
    tickUpper: int
    ## any change in liquidity
    liquidityDelta: int  

@dataclass
class SwapCache:
    ## the protocol fee for the input token
    feeProtocol: int
    ## liquidity at the beginning of the swap
    liquidityStart: int

@dataclass
class SwapState:
    ## the amount remaining to be swapped in#out of the input#output asset
    amountSpecifiedRemaining: int
    ## the amount already swapped out#in of the output#input asset
    amountCalculated: int
    ## current sqrt(price)
    sqrtPriceX96: int
    ## the tick associated with the current price
    tick: int
    ## the global fee growth of the input token
    feeGrowthGlobalX128: int
    ## amount of input token paid as protocol fee
    protocolFee: int
    ## the current liquidity in range
    liquidity: int

    ## list of ticks crossed during the swap
    ticksCrossed: list

@dataclass
class StepComputations:
    ## the price at the beginning of the step
    sqrtPriceStartX96: int
    ## the next tick to swap to from the current tick in the swap direction
    tickNext: int
    ## whether tickNext is initialized or not
    initialized: bool
    ## sqrt(price) for the next tick (1#0)
    sqrtPriceNextX96: int
    ## how much is being swapped in in this step
    amountIn: int
    ## how much is being swapped out
    amountOut: int
    ## how much fee is being paid in
    feeAmount: int

@dataclass
class ProtocolFees:
    token0: int
    token1: int

class UniswapV3Exchange(IExchange, LPERC20):

    """ 
        Uniswap V3 Exchange  

        Parameters
        -----------------
        factory_struct : FactoryInit
            Factory initialization data
        exchg_struct : UniswapExchangeInit
            Exchange initialization data           
    """       
                       
    def __init__(self, factory_struct: FactoryData, exchg_struct: UniswapExchangeData):
        super().__init__(exchg_struct.tkn0.token_name+exchg_struct.tkn1.token_name, exchg_struct.address)
        self.version = exchg_struct.version
        self.factory = factory_struct
        self.precision = exchg_struct.precision
        self.token0 = exchg_struct.tkn0.token_name     
        self.token1 = exchg_struct.tkn1.token_name       
        self.reserve0 = 0             
        self.reserve1 = 0 
        self.fee = exchg_struct.fee
        self.aggr_fee0 = 0
        self.aggr_fee1 = 0
        self.collected_fee0 = 0
        self.collected_fee1 = 0              
        self.name =  f"{self.token0}-{self.token1}"
        self.symbol = exchg_struct.symbol
        self.precision = exchg_struct.precision
        self.last_liquidity_deposit = 0
        self.total_supply = 0
        self.slot0 = Slot0(0, 0, 0)
        self.positions = {}
        self.ticks = {}
        self.feeGrowthGlobal0X128 = 0
        self.feeGrowthGlobal1X128 = 0  
        self.protocolFees = ProtocolFees(0, 0)
        self.tickSpacing = exchg_struct.tick_spacing
        self.maxLiquidityPerTick = Tick.tickSpacingToMaxLiquidityPerTick(self.tickSpacing)      

    def summary(self):

        """ summary
            Summary print-out of exchange, reserves and liquidity              
        """         
        
        tokens = self.factory.token_from_exchange[self.name] 
        x_tkn = tokens[self.token0]
        y_tkn = tokens[self.token1]          
        print(f"Exchange {self.name} ({self.symbol})")
        if (self.precision == UniswapExchangeData.TYPE_GWEI):
            print(f"Real Reserves:   {self.token0} = {self.reserve0}, {self.token1} = {self.reserve1}")
            print(f"Gross Liquidity: {self.total_supply} \n")
        else:  
            res0 = UniV3Helper().gwei2dec(self.reserve0)
            res1 = UniV3Helper().gwei2dec(self.reserve1)
            print(f"Real Reserves:   {self.token0} = {res0}, {self.token1} = {res1}")
            print(f"Gross Liquidity: {UniV3Helper().gwei2dec(self.total_supply)} \n")            

    def initialize(self, sqrtPriceX96):

        """ initialize

            Sets the initial price for the pool
                
            Parameters
            -----------------
            sqrtPriceX96 : float
                the initial sqrt price of the pool as a Q64.96                     
        """  
        
        checkInputTypes(uint160=(sqrtPriceX96))
        assert self.slot0.sqrtPriceX96 == 0, "UniswapV3: AI"

        tick = TickMath.getTickAtSqrtRatio(sqrtPriceX96)

        self.slot0 = Slot0(
            sqrtPriceX96,
            tick,
            0,
        )
    
    def mint(self, recipient, tickLower, tickUpper, amount): 

        """ mint

            Adds liquidity for the given recipient/tickLower/tickUpper position. The final amounts 
            calculated are automatically transferred from the swapper to the pool and vice verse. 
            The amount of token0/token1 due depends on tickLower, tickUpper, the amount of liquidity, 
            and the current price.
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            tickLower : int
                Lower tick of the position in which to add liquidity  
            tickUpper : int
                Upper tick of the position in which to add liquidity 
            amount : int
                Amount of liquidity to mint  
                
            Returns
            -------
            amount0 : float
                Amount of token0 that was paid to mint the given amount of liquidity.   
            amount1 : float
                Amount of token1 that was paid to mint the given amount of liquidity.                   
        """          
        amount = self.convert_to_machine(amount)
        
        checkInputTypes(
            accounts=(recipient), int24=(tickLower, tickUpper), uint128=(amount)
        )
        assert amount > 0

        (_, amount0Int, amount1Int) = self._modifyPosition(
            ModifyPositionParams(recipient, tickLower, tickUpper, amount)
        )

        amount0 = toUint256(abs(amount0Int))
        amount1 = toUint256(abs(amount1Int))
        
        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV3: TOKEN_UNAVAILABLE' 

        balance0Before = tokens.get(self.token0).token_total
        balance1Before = tokens.get(self.token1).token_total 
        
        tokens.get(self.token0).deposit(recipient, amount0)
        tokens.get(self.token1).deposit(recipient, amount1)  

        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total

        self._update(balanceA, balanceB)
    
        assert balance0Before + amount0 <= tokens.get(self.token0).token_total, 'UniswapV3: M0' 
        assert balance1Before + amount1 <= tokens.get(self.token1).token_total, 'UniswapV3: M0' 
 
        amount0 = self.convert_to_human(amount0)
        amount1 = self.convert_to_human(amount1)        
              
        return (amount0, amount1)
        

    def collect(self, recipient, tickLower, tickUpper, amount0Requested, amount1Requested):

        """ collect

            Collects tokens owed to a position. Does not recompute fees earned, which must be done either 
            via mint or burn of any amount of liquidity. Collect must be called by the position owner. 
            To withdraw only token0 or only token1, amount0Requested or amount1Requested may be set to zero. 
            To withdraw all tokens owed, caller may pass any value greater than the actual tokens owed, e.g. 
            type(uint128).max. Tokens owed may be from accumulated swap fees or burned liquidity.
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            tickLower : int
                Lower tick of the position in which to add liquidity  
            tickUpper : int
                Lower tick of the position in which to add liquidity                 
            amount0Requested : int
                How much token0 should be withdrawn from the fees owed
            amount1Requested : int
                How much token1 should be withdrawn from the fees owed  
                
            Returns
            -------
            amount0 : float
                Amount of token0 that was paid to mint the given amount of liquidity.   
            amount1 : float
                Amount of token1 that was paid to mint the given amount of liquidity.                   
        """ 

        amount0Requested = self.convert_to_machine(amount0Requested)
        amount1Requested = self.convert_to_machine(amount1Requested)
        
        checkInputTypes(
            accounts=(recipient),
            int24=(tickLower, tickUpper),
            uint128=(amount0Requested, amount1Requested),
        )
        # Add this check to prevent creating a new position if the position doesn't exist or it's empty
        position = Position.assertPositionExists(
            self.positions, recipient, tickLower, tickUpper
        )

        amount0 = (
            position.tokensOwed0
            if (amount0Requested > position.tokensOwed0)
            else amount0Requested
        )
        amount1 = (
            position.tokensOwed1
            if (amount1Requested > position.tokensOwed1)
            else amount1Requested
        )

        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0) and tokens.get(self.token1), 'UniswapV3: TOKEN_UNAVAILABLE' 
        
        if amount0 > 0:
            position.tokensOwed0 -= amount0
            tokens.get(self.token0).deposit(recipient, amount0)
            #self.ledger.transferToken(self, recipient, self.token0, amount0)
        if amount1 > 0:
            position.tokensOwed1 -= amount1
            tokens.get(self.token1).deposit(recipient, amount1) 
            #self.ledger.transferToken(self, recipient, self.token1, amount1)
  
        amount0 = self.convert_to_human(amount0)
        amount1 = self.convert_to_human(amount1)        

        return (recipient, tickLower, tickUpper, amount0, amount1)   
        

    def burn(self, recipient, tickLower, tickUpper, amount):

        """ burn

            Burn liquidity from the sender and account tokens owed for the liquidity to the position. Can 
            be used to trigger a recalculation of fees owed to a position by calling with an amount of 0. 
            Fees must be collected separately via a call to collect
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            tickLower : int
                Lower tick of the position in which to add liquidity  
            tickUpper : int
                Lower tick of the position in which to add liquidity                 
            amount : int
                How much liquidity to burn
                
            Returns
            -------
            recipient : str
                Address for which the liquidity will be created 
            amount0 : int
                Delta of the balance of token0 of the pool, exact when negative, minimum when positive
            amount1 : int
                Delta of the balance of token1 of the pool, exact when negative, minimum when positive               
            amount : int
                How much liquidity to burn                  
        """  
        amount = self.convert_to_machine(amount)
        
        checkInputTypes(
            accounts=(recipient), int24=(tickLower, tickUpper), uint128=(amount)
        )

        # Add check if the position exists - when poking an uninitialized position it can be that
        # getFeeGrowthInside finds a non-initialized tick before Position.update reverts.
        Position.assertPositionExists(self.positions, recipient, tickLower, tickUpper)

        # Added extra recipient input variable to mimic msg.sender
        (position, amount0Int, amount1Int) = self._modifyPosition(
            ModifyPositionParams(recipient, tickLower, tickUpper, -amount)
        )

        tokens = self.factory.token_from_exchange[self.name]
        tokens.get(self.token0).deposit(recipient, amount0Int)
        tokens.get(self.token1).deposit(recipient, amount1Int)     

        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total

        self._update(balanceA, balanceB)        

        # Mimic conversion to uint256
        amount0 = abs(-amount0Int) & (2**256 - 1)
        amount1 = abs(-amount1Int) & (2**256 - 1)        

        if amount0 > 0 or amount1 > 0:
            position.tokensOwed0 += amount0
            position.tokensOwed1 += amount1

        amount = self.convert_to_human(amount)   
        amount0 = self.convert_to_human(amount0)
        amount1 = self.convert_to_human(amount1)
             
        return (recipient, tickLower, tickUpper, amount, amount0, amount1)



    def swapExact0For1(self, recipient, amount, sqrtPriceLimit):

        """ swapExact0For1

            Swap exact value of token0 for token1
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            amount : int
                How much token to swap
            sqrtPriceLimit : int
                Used to determine the highest price in the swap, and needs to be set when swapping on 
                the pool directly.       
                
            Returns
            -------
            recipient : str
                Address for which the liquidity will be created 
            amount0 : int
                Delta of the balance of token0 of the pool, exact when negative, minimum when positive
            amount1 : int
                Delta of the balance of token1 of the pool, exact when negative, minimum when positive                             
        """         
        amount = self.convert_to_machine(amount)
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else UniV3Utils.getSqrtPriceLimitX96('Token0')
        )

        return self._swap(UniV3Utils.TEST_TOKENS[0], [amount, 0], recipient, sqrtPriceLimitX96)  

    def swap0ForExact1(self, recipient,  amount, sqrtPriceLimit):

        """ swapExact0For1

            Swap token0 for exact value of token1
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            amount : int
                How much token to swap
            sqrtPriceLimit : int
                Used to determine the highest price in the swap, and needs to be set when swapping 
                on the pool directly.       
                
            Returns
            -------
            recipient : str
                Address for which the liquidity will be created 
            amount0 : int
                Delta of the balance of token0 of the pool, exact when negative, minimum when positive
            amount1 : int
                Delta of the balance of token1 of the pool, exact when negative, minimum when positive                             
        """           
        
        amount = self.convert_to_machine(amount)
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else UniV3Utils.getSqrtPriceLimitX96(UniV3Utils.TEST_TOKENS[0])
        )
        return self._swap(UniV3Utils.TEST_TOKENS[0], [0, amount], recipient, sqrtPriceLimitX96)  


    def swapExact1For0(self, recipient,  amount, sqrtPriceLimit):

        """ swapExact0For1

            Swap exact value of token1 for token0
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            amount : int
                How much token to swap
            sqrtPriceLimit : int
                Used to determine the highest price in the swap, and needs to be set when swapping 
                on the pool directly.       
                
            Returns
            -------
            recipient : str
                Address for which the liquidity will be created 
            amount0 : int
                Delta of the balance of token0 of the pool, exact when negative, minimum when positive
            amount1 : int
                Delta of the balance of token1 of the pool, exact when negative, minimum when positive                             
        """          
        
        amount = amount if self.precision == UniswapExchangeData.TYPE_GWEI else UniV3Helper().dec2gwei(amount)
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else UniV3Utils.getSqrtPriceLimitX96(UniV3Utils.TEST_TOKENS[1])
        )
        return self._swap(UniV3Utils.TEST_TOKENS[1], [amount, 0], recipient, sqrtPriceLimitX96)
    
    def swap1ForExact0(self, recipient, amount, sqrtPriceLimit):

        """ swapExact0For1

            Swap token1 for exact value of token0
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            amount : int
                How much token to swap
            sqrtPriceLimit : int
                Used to determine the highest price in the swap, and needs to be set when swapping 
                on the pool directly.       
                
            Returns
            -------
            recipient : str
                Address for which the liquidity will be created 
            amount0 : int
                Delta of the balance of token0 of the pool, exact when negative, minimum when positive
            amount1 : int
                Delta of the balance of token1 of the pool, exact when negative, minimum when positive                             
        """         

        amount = self.convert_to_machine(amount)
        sqrtPriceLimitX96 = (
            sqrtPriceLimit
            if sqrtPriceLimit != None
            else UniV3Utils.getSqrtPriceLimitX96(UniV3Utils.TEST_TOKENS[1])
        )
        return self._swap(UniV3Utils.TEST_TOKENS[1], [0, amount], recipient, sqrtPriceLimitX96)     

    def swap(self, recipient, zeroForOne, amountSpecified, sqrtPriceLimitX96):

        """ swap

            Swap token0 for token1, or token1 for token0. The tokens are automatically transferred 
            at the end of the swapping function.
                
            Parameters
            -----------------    
            recipient : str
                Address for which the liquidity will be created      
            zeroForOne : int
                the direction of the swap, true for token0 to token1, false for token1 to token0 
            amountSpecified : int
                The amount of the swap, which implicitly configures the swap as exact input 
                (positive), or exact output (negative)        
            sqrtPriceLimitX96 : int
                The Q64.96 sqrt price limit. If zero for one, the price cannot be less than this 
                value after the swap. If one for zero, the price cannot be greater than this 
                value after the swap
                
            Returns
            -------
            recipient : str
                Address for which the liquidity will be created 
            amount0 : int
                Delta of the balance of token0 of the pool, exact when negative, minimum when positive
            amount1 : int
                Delta of the balance of token1 of the pool, exact when negative, minimum when positive                             
        """ 
        
        checkInputTypes(
            accounts=(recipient),
            bool=(zeroForOne),
            int256=(amountSpecified),
            uint160=(sqrtPriceLimitX96),
        )
        assert amountSpecified != 0, "UniswapV3: AS"

        slot0Start = self.slot0        
        
        if zeroForOne:
            assert (
                sqrtPriceLimitX96 < slot0Start.sqrtPriceX96
                and sqrtPriceLimitX96 > TickMath.MIN_SQRT_RATIO
            ), "UniswapV3: ZEROFORONE SPL"
        else:
            assert (
                sqrtPriceLimitX96 > slot0Start.sqrtPriceX96
                and sqrtPriceLimitX96 < TickMath.MAX_SQRT_RATIO
            ), "UniswapV3: ONEFORZERO SPL"
  
        feeProtocol = (
            (slot0Start.feeProtocol % 16)
            if zeroForOne
            else (slot0Start.feeProtocol >> 4)
        )

        cache = SwapCache(feeProtocol, self.total_supply)

        exactInput = amountSpecified > 0

        state = SwapState(
            amountSpecified,
            0,
            slot0Start.sqrtPriceX96,
            slot0Start.tick,
            self.feeGrowthGlobal0X128 if zeroForOne else self.feeGrowthGlobal1X128,
            0,
            cache.liquidityStart,
            [],
        )

        while (
            state.amountSpecifiedRemaining != 0
            and state.sqrtPriceX96 != sqrtPriceLimitX96
        ):
            step = StepComputations(0, 0, 0, 0, 0, 0, 0)
            step.sqrtPriceStartX96 = state.sqrtPriceX96

            (step.tickNext, step.initialized) = self.nextTick(state.tick, zeroForOne)

            ## get the price for the next tick
            step.sqrtPriceNextX96 = TickMath.getSqrtRatioAtTick(step.tickNext)

            ## compute values to swap to the target tick, price limit, or point where input#output amount is exhausted
            if zeroForOne:
                sqrtRatioTargetX96 = (
                    sqrtPriceLimitX96
                    if step.sqrtPriceNextX96 < sqrtPriceLimitX96
                    else step.sqrtPriceNextX96
                )
            else:
                sqrtRatioTargetX96 = (
                    sqrtPriceLimitX96
                    if step.sqrtPriceNextX96 > sqrtPriceLimitX96
                    else step.sqrtPriceNextX96
                )

            (
                state.sqrtPriceX96,
                step.amountIn,
                step.amountOut,
                step.feeAmount,
            ) = SwapMath.computeSwapStep(
                state.sqrtPriceX96,
                sqrtRatioTargetX96,
                state.liquidity,
                state.amountSpecifiedRemaining,
                self.fee,
            )

            if exactInput:
                state.amountSpecifiedRemaining -= step.amountIn + step.feeAmount
                state.amountCalculated = SafeMath.subInts(
                    state.amountCalculated, step.amountOut
                )
            else:
                state.amountSpecifiedRemaining += step.amountOut
                state.amountCalculated = SafeMath.addInts(
                    state.amountCalculated, step.amountIn + step.feeAmount
                )

            ## if the protocol fee is on, calculate how much is owed, decrement feeAmount, and increment protocolFee
            if cache.feeProtocol > 0:
                delta = abs(step.feeAmount // cache.feeProtocol)
                step.feeAmount -= delta
                state.protocolFee += delta & (2**128 - 1)

            ## update global fee tracker
            if state.liquidity > 0:
                state.feeGrowthGlobalX128 += FullMath.mulDiv(
                    step.feeAmount, FixedPoint128_Q128, state.liquidity
                )
                # Addition can overflow in Solidity - mimic it
                state.feeGrowthGlobalX128 = toUint256(state.feeGrowthGlobalX128)

            ## shift tick if we reached the next price
            if state.sqrtPriceX96 == step.sqrtPriceNextX96:
                ## if the tick is initialized, run the tick transition
                ## @dev: here is where we should handle the case of an uninitialized boundary tick
                if step.initialized:
                    liquidityNet = Tick.cross(
                        self.ticks,
                        step.tickNext,
                        state.feeGrowthGlobalX128
                        if zeroForOne
                        else self.feeGrowthGlobal0X128,
                        self.feeGrowthGlobal1X128
                        if zeroForOne
                        else state.feeGrowthGlobalX128,
                    )
                    ## if we're moving leftward, we interpret liquidityNet as the opposite sign
                    ## safe because liquidityNet cannot be type(int128).min
                    if zeroForOne:
                        liquidityNet = -liquidityNet

                    state.liquidity = LiquidityMath.addDelta(
                        state.liquidity, liquidityNet
                    )

                state.tick = (step.tickNext - 1) if zeroForOne else step.tickNext
            elif state.sqrtPriceX96 != step.sqrtPriceStartX96:
                ## recompute unless we're on a lower tick boundary (i.e. already transitioned ticks), and haven't moved
                state.tick = TickMath.getTickAtSqrtRatio(state.sqrtPriceX96)

        ## End of swap loop
        ## update tick
        if state.tick != slot0Start.tick:
            self.slot0.sqrtPriceX96 = state.sqrtPriceX96
            self.slot0.tick = state.tick
        else:
            ## otherwise just update the price
            self.slot0.sqrtPriceX96 = state.sqrtPriceX96

        ## update liquidity if it changed
        if cache.liquidityStart != state.liquidity:
            self.liquidity = state.liquidity

        ## update fee growth global and, if necessary, protocol fees
        ## overflow is acceptable, protocol has to withdraw before it hits type(uint128).max fees

        if zeroForOne:
            self.feeGrowthGlobal0X128 = state.feeGrowthGlobalX128
            if state.protocolFee > 0:
                self.protocolFees.token0 += state.protocolFee
        else:
            self.feeGrowthGlobal1X128 = state.feeGrowthGlobalX128
            if state.protocolFee > 0:
                self.protocolFees.token1 += state.protocolFee

        (amount0, amount1) = (
            (amountSpecified - state.amountSpecifiedRemaining, state.amountCalculated)
            if (zeroForOne == exactInput)
            else (
                state.amountCalculated,
                amountSpecified - state.amountSpecifiedRemaining,
            )
        )
        
        tokens = self.factory.token_from_exchange[self.name]
        if zeroForOne: 
            tokens.get(self.token0).deposit(recipient, abs(amount0))
            self._swap_tokens(0, abs(amount1), recipient)            
        else: 
            tokens.get(self.token1).deposit(recipient, abs(amount1))
            self._swap_tokens(abs(amount0), 0, recipient)            

        amount0 = self.convert_to_human(amount0)
        amount1 = self.convert_to_human(amount1)
        liquidity = self.convert_to_human(state.liquidity)
        self._update_fees()
        
        return (
            recipient,
            amount0,
            amount1,
            state.sqrtPriceX96,
            liquidity,
            state.tick,
        )

    def setFeeProtocol(self, feeProtocol0, feeProtocol1):

        """ setFeeProtocol

            Set the denominator of the protocol's % share of the fees
            
            Parameters
            -----------------    
            feeProtocol0 : int
                New protocol fee for token0 of the pool      
            feeProtocol1 : int
                New protocol fee for token1 of the pool                        
        """ 
        
        checkInputTypes(uint8=(feeProtocol0, feeProtocol1))
        assert (feeProtocol0 == 0 or (feeProtocol0 >= 4 and feeProtocol0 <= 10)) and (
            feeProtocol1 == 0 or (feeProtocol1 >= 4 and feeProtocol1 <= 10)
        )

        feeProtocolOld = self.slot0.feeProtocol
        feeProtocolNew = feeProtocol0 + (feeProtocol1 << 4)
        # Health check
        checkUInt8(feeProtocolNew)
        self.slot0.feeProtocol = feeProtocolNew
        return (feeProtocolOld % 16, feeProtocolOld >> 4, feeProtocol0, feeProtocol1)

    def checkTicks(self, tickLower, tickUpper):

        """ checkTicks

            Common checks for valid tick inputs
            
            Parameters
            -----------------    
            tickLower : int
                Lower tick of the position in which to add liquidity  
            tickUpper : int
                Lower tick of the position in which to add liquidity                    
        """ 
        
        checkInputTypes(int24=(tickLower, tickUpper))
        assert tickLower < tickUpper, "UniswapV3: TLU"
        assert tickLower >= TickMath.MIN_TICK, "UniswapV3: TLM"
        assert tickUpper <= TickMath.MAX_TICK, "UniswapV3: TUM"        

    def nextTick(self, tick, lte):

        """ nextTick

            It is assumed that the keys are within [MIN_TICK , MAX_TICK], which should always be the 
            case. We don't run the risk of overshooting tickNext (out of boundaries) as long as ticks 
            (keys) have been initialized within the boundaries. However, if there is no initialized 
            tick to the left or right we will return the next boundary. Then we need to return the 
            initialized bool to indicate that we are at the boundary and it is not an initalized tick.
            
            Parameters
            -----------------    
            tick : int
                The starting tick    
            lte : int
                Whether to search for the next initialized tick to the left (less than or equal to the starting tick)
                
            Returns
            -------
            nextTick : int
                Next tick with liquidity to be used in the swap function                      
        """ 
        
        checkInputTypes(int24=(tick), bool=(lte))

        keyList = list(self.ticks.keys())

        # If tick doesn't exist in the mapping we fake it (easier than searching for nearest value). This is probably not the
        # best way, but it is a simple and intuitive way to reproduce the behaviour of the logic.
        if not self.ticks.__contains__(tick):
            keyList += [tick]
        sortedKeyList = sorted(keyList)
        indexCurrentTick = sortedKeyList.index(tick)

        if lte:
            # If the current tick is initialized (not faked), we return the current tick
            if self.ticks.__contains__(tick):
                return tick, True
            elif indexCurrentTick == 0:
                # No tick to the left
                return TickMath.MIN_TICK, False
            else:
                nextTick = sortedKeyList[indexCurrentTick - 1]
        else:

            if indexCurrentTick == len(sortedKeyList) - 1:
                # No tick to the right
                return TickMath.MAX_TICK, False
            nextTick = sortedKeyList[indexCurrentTick + 1]

        # Return tick within the boundaries
        return nextTick, True  

    def get_price(self, token):  
        
        """ get_price

            Get price of select token in the exchange pair
                
            Parameters
            -----------------
            token : ERC20
                ERC20 token                
        """          
        sqrt_P = self.slot0.sqrtPriceX96/2**96
        
        if(token.token_name == self.token0):
            if(self.reserve0 == 0):
                return None 
            else:
                return sqrt_P**2 
        elif(token.token_name == self.token1):
            if(self.reserve1 == 0):
                return None
            else:
                return 1/sqrt_P**2 
        else:
            assert False, 'UniswapV2: WRONG_INPUT_TOKEN'   


    def get_last_liquidity_deposit(self):  
        
        """ get_last_liquidity_deposit

            Get the last liquidity deposit that went into pool         
        """          

        return self.last_liquidity_deposit    


    def get_liquidity_from_provider(self, provider_account):  
        
        """ get_liquidity_from_provider

            Get liquidity from provider  
            
            Parameters
            -----------------
            provider_account : str
                Provider account name  
            
        """          

        return self.convert_to_human(self.liquidity_providers[provider_account])     
    
                  
    def get_liquidity(self):  
        
        """ get_liquidity

            Get liquidity of exchange pool         
        """          

        return self.convert_to_human(self.total_supply)         
            
    def get_reserve(self, token):  
        
        """ get_reserve

            Get reserve amount of select token in the exchange pair
                
            Parameters
            -----------------
            token : ERC20
                ERC20 token                
        """         
        
        if(token.token_name == self.token0):
            return self.convert_to_human(self.reserve0)
        elif(token.token_name == self.token1):
            return self.convert_to_human(self.reserve1) 
        else:
            assert False, 'UniswapV3: WRONG_INPUT_TOKEN'      

    def get_virtual_reserve(self, token):  
        
        """ get_virtual_reserve

            Get virtual reserve amount of select token in the exchange pair
                
            Parameters
            -----------------
            token : ERC20
                ERC20 token                
        """         
        
        sqrt_P = self.slot0.sqrtPriceX96/2**96
        liq = self.get_liquidity()
        
        if(token.token_name == self.token0):
            return liq/sqrt_P
        elif(token.token_name == self.token1):
            return liq*sqrt_P
        else:
            assert False, 'UniswapV3: WRONG_INPUT_TOKEN'           

    def convert_to_human(self, val): 
        val = val if self.precision == UniswapExchangeData.TYPE_GWEI else UniV3Helper().gwei2dec(val)
        return val

    def convert_to_machine(self, val): 
        val = int(val) if self.precision == UniswapExchangeData.TYPE_GWEI else UniV3Helper().dec2gwei(val)
        return val   

    def _update_fees(self): 
        liquidity = UniV3Helper().gwei2dec(self.total_supply)
        self.collected_fee0 = liquidity*self.feeGrowthGlobal0X128/2**128
        self.collected_fee1 = liquidity*self.feeGrowthGlobal1X128/2**128
    
    def _swap(self, inputToken, amounts, recipient, sqrtPriceLimitX96):
        [amountIn, amountOut] = amounts
        exactInput = amountOut == 0
        amount = amountIn if exactInput else amountOut

        if inputToken == 'Token0':
            if exactInput:
                checkInt128(amount)
                return self.swap(recipient, True, amount, sqrtPriceLimitX96)
            else:
                checkInt128(-amount)
                return self.swap(recipient, True, -amount, sqrtPriceLimitX96)
        else:
            if exactInput:
                checkInt128(amount)
                return self.swap(recipient, False, amount, sqrtPriceLimitX96)                  
            else:
                checkInt128(-amount)
                return self.swap(recipient, False, -amount, sqrtPriceLimitX96)      
    
    def _swap_tokens(self, amountA_out, amountB_out, to_addr):
        
        """ _swap_tokens

            Remove liquidity from both coins in the pair based on lp amount
                
            Parameters
            -----------------
            amountA_out : float
                Swap amountA out
            amountB_out : float
                Swap amountB out               
            to_addr : str
                Receiving user address                   
        """         
        
        assert amountA_out > 0 or amountB_out > 0, 'UniswapV3: INSUFFICIENT_OUTPUT_AMOUNT'
        assert amountA_out <= self.reserve0 and amountB_out <= self.reserve1, 'UniswapV3: INSUFFICIENT_LIQUIDITY'

        tokens = self.factory.token_from_exchange[self.name]
        assert tokens.get(self.token0).token_addr != to_addr, 'UniswapV3: INVALID_TO_ADDRESS'
        assert tokens.get(self.token1).token_addr != to_addr, 'UniswapV3: INVALID_TO_ADDRESS'
        
        tokens.get(self.token0).transfer(to_addr, amountA_out)
        tokens.get(self.token1).transfer(to_addr, amountB_out)    
        
        balanceA = tokens.get(self.token0).token_total
        balanceB = tokens.get(self.token1).token_total

        amountA_in = balanceA - (self.reserve0 - amountA_out) if balanceA > self.reserve0 - amountA_out else 0
        amountB_in = balanceB - (self.reserve1 - amountB_out) if balanceB > self.reserve1 - amountB_out else 0

        assert amountA_in > 0 or amountB_in > 0, 'UniswapV3: INSUFFICIENT_INPUT_AMOUNT'
    
        self._update(balanceA, balanceB)    
    


    def _update(self, balanceA, balanceB):
        
        """ _update

            Update reserve amounts for both coins in the pair
                
            Parameters
            -----------------   
            balanceA : float
                New reserve amount of A      
            balance1 : float
                New reserve amount of B                   
        """         
        
        self.reserve0 = balanceA
        self.reserve1 = balanceB    
    
    def _modifyPosition(self, params):

        checkInputTypes(
            accounts=(params.owner),
            int24=(params.tickLower, params.tickUpper),
            int128=(params.liquidityDelta),
        )
        self.checkTicks(params.tickLower, params.tickUpper)

        # Initialize values
        amount0 = amount1 = 0

        position = self._updatePosition(
            params.owner,
            params.tickLower,
            params.tickUpper,
            params.liquidityDelta,
            self.slot0.tick,
        )

        if params.liquidityDelta != 0:
            if self.slot0.tick < params.tickLower:
                ## current tick is below the passed range; liquidity can only become in range by crossing from left to
                ## right, when we'll need _more_ token0 (it's becoming more valuable) so user must provide it
                amount0 = SqrtPriceMath.getAmount0DeltaHelper(
                    TickMath.getSqrtRatioAtTick(params.tickLower),
                    TickMath.getSqrtRatioAtTick(params.tickUpper),
                    params.liquidityDelta,
                )
            elif self.slot0.tick < params.tickUpper:
                ## current tick is inside the passed range
                amount0 = SqrtPriceMath.getAmount0DeltaHelper(
                    self.slot0.sqrtPriceX96,
                    TickMath.getSqrtRatioAtTick(params.tickUpper),
                    params.liquidityDelta,
                )
                amount1 = SqrtPriceMath.getAmount1DeltaHelper(
                    TickMath.getSqrtRatioAtTick(params.tickLower),
                    self.slot0.sqrtPriceX96,
                    params.liquidityDelta,
                )
            else:
                ## current tick is above the passed range; liquidity can only become in range by crossing from right to
                ## left, when we'll need _more_ token1 (it's becoming more valuable) so user must provide it
                amount1 = SqrtPriceMath.getAmount1DeltaHelper(
                    TickMath.getSqrtRatioAtTick(params.tickLower),
                    TickMath.getSqrtRatioAtTick(params.tickUpper),
                    params.liquidityDelta,
                )
            self.total_supply = LiquidityMath.addDelta(
                self.total_supply, params.liquidityDelta
            )
            self.last_liquidity_deposit = self.convert_to_human(params.liquidityDelta)

        return (position, amount0, amount1)  
    
    def _updatePosition(self, owner, tickLower, tickUpper, liquidityDelta, tick):
        checkInputTypes(
            accounts=(owner),
            int24=(tickLower, tickUpper, tick),
            int128=(liquidityDelta),
        )
        # This will create a position if it doesn't exist

        
        position = Position.get(self.positions, owner, tickLower, tickUpper)

        # Initialize values
        flippedLower = flippedUpper = False

        ## if we need to update the ticks, do it
        if liquidityDelta != 0:
            flippedLower = Tick.update(
                self.ticks,
                tickLower,
                tick,
                liquidityDelta,
                self.feeGrowthGlobal0X128,
                self.feeGrowthGlobal1X128,
                False,
                self.maxLiquidityPerTick,
            )
            flippedUpper = Tick.update(
                self.ticks,
                tickUpper,
                tick,
                liquidityDelta,
                self.feeGrowthGlobal0X128,
                self.feeGrowthGlobal1X128,
                True,
                self.maxLiquidityPerTick,
            )

        if flippedLower:
            assert tickLower % self.tickSpacing == 0  ## ensure that the tick is spaced
        if flippedUpper:
            assert tickUpper % self.tickSpacing == 0  ## ensure that the tick is spaced

        (feeGrowthInside0X128, feeGrowthInside1X128) = Tick.getFeeGrowthInside(
            self.ticks,
            tickLower,
            tickUpper,
            tick,
            self.feeGrowthGlobal0X128,
            self.feeGrowthGlobal1X128,
        )

        Position.update(
            position, liquidityDelta, feeGrowthInside0X128, feeGrowthInside1X128
        )

        ## clear any tick data that is no longer needed
        if liquidityDelta < 0:
            if flippedLower:
                Tick.clear(self.ticks, tickLower)
            if flippedUpper:
                Tick.clear(self.ticks, tickUpper)
        return position    
    



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

# Original MIT licenced TickMath file from chainflip-io 
# - https://github.com/chainflip-io/chainflip-uniswapV3-python

from .Shared import *

### @notice Calculates sqrt(1.0001^tick) * 2^96
### @dev Throws if |tick| > max tick
### @param tick The input tick for the above formula
### @return sqrtPriceX96 A Fixed point Q64.96 number representing the sqrt of the ratio of the two assets (token1/token0)
### at the given tick
def getSqrtRatioAtTick(tick):
    checkInt24(tick)
    absTick = abs(tick)
    assert absTick <= MAX_TICK, "T"

    ratio = (
        0xFFFCB933BD6FAD37AA2D162D1A594001
        if absTick & 0x1 != 0
        else 0x100000000000000000000000000000000
    )

    if absTick & 0x2 != 0:
        ratio = (ratio * 0xFFF97272373D413259A46990580E213A) >> 128
    if absTick & 0x4 != 0:
        ratio = (ratio * 0xFFF2E50F5F656932EF12357CF3C7FDCC) >> 128
    if absTick & 0x8 != 0:
        ratio = (ratio * 0xFFE5CACA7E10E4E61C3624EAA0941CD0) >> 128
    if absTick & 0x10 != 0:
        ratio = (ratio * 0xFFCB9843D60F6159C9DB58835C926644) >> 128
    if absTick & 0x20 != 0:
        ratio = (ratio * 0xFF973B41FA98C081472E6896DFB254C0) >> 128
    if absTick & 0x40 != 0:
        ratio = (ratio * 0xFF2EA16466C96A3843EC78B326B52861) >> 128
    if absTick & 0x80 != 0:
        ratio = (ratio * 0xFE5DEE046A99A2A811C461F1969C3053) >> 128
    if absTick & 0x100 != 0:
        ratio = (ratio * 0xFCBE86C7900A88AEDCFFC83B479AA3A4) >> 128
    if absTick & 0x200 != 0:
        ratio = (ratio * 0xF987A7253AC413176F2B074CF7815E54) >> 128
    if absTick & 0x400 != 0:
        ratio = (ratio * 0xF3392B0822B70005940C7A398E4B70F3) >> 128
    if absTick & 0x800 != 0:
        ratio = (ratio * 0xE7159475A2C29B7443B29C7FA6E889D9) >> 128
    if absTick & 0x1000 != 0:
        ratio = (ratio * 0xD097F3BDFD2022B8845AD8F792AA5825) >> 128
    if absTick & 0x2000 != 0:
        ratio = (ratio * 0xA9F746462D870FDF8A65DC1F90E061E5) >> 128
    if absTick & 0x4000 != 0:
        ratio = (ratio * 0x70D869A156D2A1B890BB3DF62BAF32F7) >> 128
    if absTick & 0x8000 != 0:
        ratio = (ratio * 0x31BE135F97D08FD981231505542FCFA6) >> 128
    if absTick & 0x10000 != 0:
        ratio = (ratio * 0x9AA508B5B7A84E1C677DE54F3E99BC9) >> 128
    if absTick & 0x20000 != 0:
        ratio = (ratio * 0x5D6AF8DEDB81196699C329225EE604) >> 128
    if absTick & 0x40000 != 0:
        ratio = (ratio * 0x2216E584F5FA1EA926041BEDFE98) >> 128
    if absTick & 0x80000 != 0:
        ratio = (ratio * 0x48A170391F7DC42444E8FA2) >> 128

    if tick > 0:
        ratio = MAX_UINT256 // ratio

    ## this divides by 1<<32 rounding up to go from a Q128.128 to a Q128.96.
    ## we then downcast because we know the result always fits within 160 bits due to our tick input constraint
    ## we round up in the division so getTickAtSqrtRatio of the output price is always consistent

    remainder = 1 if ratio % 2**32 != 0 else 0
    # For some reason doing the division rounding up doesn't give the exact number
    result = (ratio >> 32) + remainder
    checkUInt160(result)
    return result


### @notice Calculates the greatest tick value such that getRatioAtTick(tick) <= ratio
### @dev Throws in case sqrtPriceX96 < MIN_SQRT_RATIO, as MIN_SQRT_RATIO is the lowest value getRatioAtTick may
### ever return.
### @param sqrtPriceX96 The sqrt ratio for which to compute the tick as a Q64.96
### @return tick The greatest tick for which the ratio is less than or equal to the input ratio
def getTickAtSqrtRatio(sqrtPriceX96):
    checkUInt160(sqrtPriceX96)
    ## second inequality must be < because the price can never reach the price at the max tick
    assert sqrtPriceX96 >= MIN_SQRT_RATIO and sqrtPriceX96 < MAX_SQRT_RATIO, "R"
    ratio = sqrtPriceX96 << 32

    r = ratio
    msb = 0

    (r, msb) = add_bit_to_log_2(r, msb, 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF, 7)
    (r, msb) = add_bit_to_log_2(r, msb, 0xFFFFFFFFFFFFFFFF, 6)
    (r, msb) = add_bit_to_log_2(r, msb, 0xFFFFFFFF, 5)
    (r, msb) = add_bit_to_log_2(r, msb, 0xFFFF, 4)
    (r, msb) = add_bit_to_log_2(r, msb, 0xFF, 3)
    (r, msb) = add_bit_to_log_2(r, msb, 0xF, 2)
    (r, msb) = add_bit_to_log_2(r, msb, 0x3, 1)
    (r, msb) = add_bit_to_log_2(r, msb, 0x1, 0)

    r = ratio >> (msb - 127) if (msb >= 128) else ratio << (127 - msb)

    ## Log2 is int in Solidity (not uint) - can be negative
    log_2 = (msb - 128) << 64

    (r, log_2) = add_fractional_bit(r, log_2, 63)
    (r, log_2) = add_fractional_bit(r, log_2, 62)
    (r, log_2) = add_fractional_bit(r, log_2, 61)
    (r, log_2) = add_fractional_bit(r, log_2, 60)
    (r, log_2) = add_fractional_bit(r, log_2, 59)
    (r, log_2) = add_fractional_bit(r, log_2, 58)
    (r, log_2) = add_fractional_bit(r, log_2, 57)
    (r, log_2) = add_fractional_bit(r, log_2, 56)
    (r, log_2) = add_fractional_bit(r, log_2, 55)
    (r, log_2) = add_fractional_bit(r, log_2, 54)
    (r, log_2) = add_fractional_bit(r, log_2, 53)
    (r, log_2) = add_fractional_bit(r, log_2, 52)
    (r, log_2) = add_fractional_bit(r, log_2, 51)
    (r, log_2) = add_fractional_bit(r, log_2, 50)

    log_sqrt10001 = log_2 * 255738958999603826347141
    ## 128.128 number

    ## TickLow and TickHi should be 24 bits long (int24)
    tickLow = (log_sqrt10001 - 3402992956809132418596140100660247210) >> 128
    tickHi = (log_sqrt10001 + 291339464771989622907027621153398088495) >> 128

    ## Add checks to ensure that the tick is in the correct lenght (<= 24 bits)
    assert tickLow >= MIN_INT24 and tickLow <= MAX_INT24, "Failure"

    if tickLow == tickHi:
        tick = tickLow
    else:
        if getSqrtRatioAtTick(tickHi) <= sqrtPriceX96:
            tick = tickHi
        else:
            tick = tickLow

    return tick


# Need to return r and msb since ints are passed by value and not by reference
def add_bit_to_log_2(r, msb, lower_bit_mask, bit):
    gt = 1 if r > lower_bit_mask else 0
    f = gt << bit
    msb = msb | f
    r = r >> f
    return (r, msb)


# Need to return r and log_2 since ints are passed by value and not by reference
def add_fractional_bit(r, log_2, bit):
    r = (r * r) >> 127
    f = r >> 128
    log_2 = log_2 | (f << bit)
    # Difference in calculation when bit == 50
    if bit != 50:
        r = r >> f
    return (r, log_2)



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import math
from decimal import Decimal
from . import TickMath

Q96 = 2**96
GWEI_PRECISION = 18

class UniV3Helper():

    
    """ Uniswap V3 helper functions             
    """       

    def __init__(self):
        pass

    def quote(self, lp, token_in, amt_tkn, lwr_tick, upr_tick):
        
        fee = 997 
        L = lp.get_liquidity()
        if(token_in.token_name == lp.token0):
            sqrtp_cur = Q96/lp.slot0.sqrtPriceX96
            sqrtp_pa = Q96/TickMath.getSqrtRatioAtTick(lwr_tick)
            sqrtp_pb = Q96/TickMath.getSqrtRatioAtTick(upr_tick)       
        elif(token_in.token_name == lp.token1):
            sqrtp_cur = lp.slot0.sqrtPriceX96/Q96
            sqrtp_pa = TickMath.getSqrtRatioAtTick(lwr_tick)/Q96
            sqrtp_pb = TickMath.getSqrtRatioAtTick(upr_tick)/Q96 
     
        sqrtp_next = sqrtp_cur + (fee*amt_tkn) / (L*1000)
        return L * (1/sqrtp_cur - 1/sqrtp_next), sqrtp_next    


    #def get_price_tick(self, human_price):  
    #    return math.floor(math.log((human_price)**2)/math.log(1.0001))   
    
    def get_price_tick(self, lp, pos = 0, human_price = None, tick_space = None):  

        """ get_tick_price
            Get tick price of select token in the exchange pair             
        """          
               
        sqrtPriceX96 = lp.slot0.sqrtPriceX96        
        human_price = (sqrtPriceX96/Q96)**2 if human_price == None else human_price
        tick_p = math.floor(math.log(human_price)/math.log(1.0001))

        tick_space = lp.tickSpacing if tick_space == None else tick_space
        if pos == -1:
            tick_p = tick_p - tick_space
        elif pos == 1:
            tick_p = tick_p + tick_space

        R = tick_p % lp.tickSpacing
        if R > int(lp.tickSpacing/2):
            tick_p += lp.tickSpacing-R
        else:
            tick_p -= R        
 
        return tick_p   
    
    def price_to_tick(self, p):
        return math.floor(math.log(p, 1.0001))    

    def tick_to_price(self, tick):
        return 1.0001**tick   

    def price_to_sqrtp(self, p):
        return int(math.sqrt(p) * Q96)    

    def sqrtp_to_price(self, sqrtp):
        #return sqrtp/(2**99)  
        return sqrtp/Q96  

    def dec2gwei(self, tkn_amt, precision=None):
        precision = GWEI_PRECISION if precision == None else precision
        return int(Decimal(str(tkn_amt))*Decimal(str(10**precision)))
    
    def gwei2dec(self, tkn_amt, precision=None):   
        precision = GWEI_PRECISION if precision == None else precision
        return float(Decimal(str(tkn_amt))/Decimal(str(10**precision)))      

    # see https://atiselsts.github.io/pdfs/uniswap-v3-liquidity-math.pdf
    def calc_Lx(self, p_sqrt_human, dx, lwr_tick, upr_tick):
        pa_sqrt_human = TickMath.getSqrtRatioAtTick(lwr_tick)/Q96
        pb_sqrt_human = TickMath.getSqrtRatioAtTick(upr_tick)/Q96
        assert p_sqrt_human <= pb_sqrt_human, "OOP"
        Lx = dx/(1/max(p_sqrt_human, pa_sqrt_human) - 1/pb_sqrt_human)
        return Lx
    
    def calc_Ly(self, p_sqrt_human, dy, lwr_tick, upr_tick, price_tick=None):
        pa_sqrt_human = TickMath.getSqrtRatioAtTick(lwr_tick)/Q96
        pb_sqrt_human = TickMath.getSqrtRatioAtTick(upr_tick)/Q96
        assert p_sqrt_human >= pa_sqrt_human, "OOP"
        Ly = dy/(min(p_sqrt_human, pb_sqrt_human) - pa_sqrt_human)
        return Ly



from decimal import *

MAX_UINT256 = 2**256 - 1

class SaferMath():

    def __init__(self):
        pass
    
    def add_div_round(self, x, y, z):
        val = self.div_round(self.add(x, y), z)
        return int(val)
    
    def add_div(self, x, y, z):
        val = (Decimal(str(x))+Decimal(str(y)))/Decimal(str(z))
        return int(val)
    
    def mul_div(self, x, y, z):
        val = (Decimal(str(x))*Decimal(str(y)))/Decimal(str(z))
        return int(val)
    
    def exp(self, x, y):
        val = Decimal(str(x)) ** Decimal(str(y))
        return int(val)
    
    def div(self, x, y):
        y = y if y != 0 else 1
        val = Decimal(str(x))/Decimal(str(y))
        return int(val)

    def sub(self, x, y):
        val = Decimal(str(x))-Decimal(str(y))
        return int(val)
    
    def add(self, x, y):
        val = Decimal(str(x))+Decimal(str(y))
        return int(val)
        
    def mul(self, x, y):
        val = Decimal(str(x))*Decimal(str(y))
        return int(val)

    def div_round(self, x, y):
        try:
            y = y if y != 0 else 1
            result = Decimal(str(x))//Decimal(str(y))        
            if Decimal(str(x)) % Decimal(str(y))  > 0:
                result += 1
            self.check_uint_256(int(result))
        except: 
            result = self.div(x,y)
            
        return int(result)

    def mul_div_round(self, x, y, z):
        try:
            # if(z == 0):
            #     z = 1
            #     x = x - 1
            out = self.div_round(self.mul(x, y), z)
        except:
            # if(z == 0):
            #     z = 1
            #     x = x - 1
            out = self.div(self.mul(x, y), z)
        return out
    
    def check_uint_256(self, val):
        assert val >= 0 and val <= MAX_UINT256, "OF or UF of UINT256"
        assert type(val) == int, "Not an integer"